#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
from unittest.mock import Mock, call, patch

import pytest

from paleomix.atomiccmd.builder import (
    AtomicCmdBuilder,
    AtomicCmdBuilderError,
    AtomicJavaCmdBuilder,
    AtomicMPICmdBuilder,
    apply_options,
)

###############################################################################
###############################################################################
# AtomicCmdBuilder: Constructor


def test_builder__simple__call():
    builder = AtomicCmdBuilder(["ls"])
    assert builder.call == ["ls"]


def test_builder__simple__str():
    builder = AtomicCmdBuilder("ls")
    assert builder.call == ["ls"]


def test_builder__simple__iterable():
    builder = AtomicCmdBuilder(iter(["ls"]))
    assert builder.call == ["ls"]


def test_builder__complex():
    builder = AtomicCmdBuilder(["java", "jar", "/a/jar"])
    assert builder.call == ["java", "jar", "/a/jar"]


def test_builder__kwargs__empty():
    builder = AtomicCmdBuilder(["ls"])
    assert builder.kwargs == {}


def test_builder__kwargs():
    expected = {"IN_FILE": "/abc/def.txt", "OUT_FILE": "/etc/fstab"}
    builder = AtomicCmdBuilder(["ls"], **expected)
    assert builder.kwargs == expected


def test_builder__kwargs__set_cwd():
    builder = AtomicCmdBuilder(["ls"], set_cwd=True)
    assert builder.kwargs == {"set_cwd": True}


###############################################################################
###############################################################################
# AtomicCmdBuilder: set_option


def test_builder__set_option():
    builder = AtomicCmdBuilder("find")
    builder.set_option("-name", "*.txt")
    assert builder.call == ["find", "-name", "*.txt"]


def test_builder__set_option__overwrite():
    builder = AtomicCmdBuilder("find")
    builder.set_option("-name", "*.txt", fixed=False)
    builder.set_option("-name", "*.bat")
    assert builder.call == ["find", "-name", "*.bat"]


def test_builder__set_option__overwrite_fixed():
    builder = AtomicCmdBuilder("find")
    builder.set_option("-name", "*.txt")
    with pytest.raises(AtomicCmdBuilderError):
        builder.set_option("-name", "*.bat")


###############################################################################
###############################################################################
# AtomicCmdBuilder: add_option


def test_builder__add_option():
    builder = AtomicCmdBuilder("find")
    builder.add_option("-name", "*.txt")
    assert builder.call == ["find", "-name", "*.txt"]


def test_builder__add_option__overwrite():
    builder = AtomicCmdBuilder("find")
    builder.add_option("-name", "*.txt")
    builder.add_option("-or")
    builder.add_option("-name", "*.bat")
    assert builder.call == ["find", "-name", "*.txt", "-or", "-name", "*.bat"]


###############################################################################
###############################################################################
# AtomicCmdBuilder: add_option / set_option common tests

_ADD_SET_OPTION = [AtomicCmdBuilder.add_option, AtomicCmdBuilder.set_option]


@pytest.mark.parametrize("setter", _ADD_SET_OPTION)
def test_builder__add_or_set_option__without_value(setter):
    builder = AtomicCmdBuilder("find")
    setter(builder, "-delete")
    assert builder.call == ["find", "-delete"]


@pytest.mark.parametrize("setter", _ADD_SET_OPTION)
def test_builder__add_or_set_option__with_sep(setter):
    builder = AtomicCmdBuilder("find")
    setter(builder, "-size", "0", sep="=")
    assert builder.call == ["find", "-size=0"]


@pytest.mark.parametrize("setter", _ADD_SET_OPTION)
def test_builder__add_or_set_option__with_non_str_value(setter):
    builder = AtomicCmdBuilder("find")
    setter(builder, "-size", 0)
    assert builder.call == ["find", "-size", 0]


def test_builder__add_or_set_option__add_and_set():
    builder = AtomicCmdBuilder("find")
    builder.set_option("-name", "*.txt")
    with pytest.raises(AtomicCmdBuilderError):
        AtomicCmdBuilder.add_option(builder, "-name", "*.bat")


def test_builder__add_or_set_option__set_and_add():
    builder = AtomicCmdBuilder("find")
    builder.add_option("-name", "*.txt")
    with pytest.raises(AtomicCmdBuilderError):
        AtomicCmdBuilder.set_option(builder, "-name", "*.bat")


@pytest.mark.parametrize("setter", _ADD_SET_OPTION)
def test_builder__add_or_set_option__with_non_str_key(setter):
    builder = AtomicCmdBuilder("find")
    with pytest.raises(TypeError):
        setter(builder, 7913, "True")


@pytest.mark.parametrize("setter", _ADD_SET_OPTION)
def test_builder__add_or_set_option__after_finalize(setter):
    builder = AtomicCmdBuilder("find")
    builder.finalize()
    with pytest.raises(AtomicCmdBuilderError):
        setter(builder, "-size", "1")


@pytest.mark.parametrize("setter", _ADD_SET_OPTION)
def test_builder__add_or_set_option__empty_key(setter):
    builder = AtomicCmdBuilder("find")
    with pytest.raises(KeyError):
        setter(builder, "", "1")


###############################################################################
###############################################################################
# AtomicCmdBuilder: pop_option


@pytest.mark.parametrize("setter", _ADD_SET_OPTION)
def test_builder__pop_option(setter):
    builder = AtomicCmdBuilder("find")
    setter(builder, "-size", "0", fixed=False)
    builder.pop_option("-size")
    assert builder.call == ["find"]


def test_builder__pop_option__last_option():
    builder = AtomicCmdBuilder("find")
    builder.add_option("-size", "0", fixed=False)
    builder.add_option("-size", "1", fixed=False)
    builder.pop_option("-size")
    assert builder.call == ["find", "-size", "0"]


@pytest.mark.parametrize("setter", _ADD_SET_OPTION)
def test_builder__pop_option__different_options(setter):
    builder = AtomicCmdBuilder("find")
    setter(builder, "-empty", fixed=False)
    setter(builder, "-size", "1", fixed=False)
    setter(builder, "-name", "*.txt", fixed=False)
    builder.pop_option("-size")
    assert builder.call == ["find", "-empty", "-name", "*.txt"]


@pytest.mark.parametrize("setter", _ADD_SET_OPTION)
def test_builder__pop_option__is_fixed(setter):
    builder = AtomicCmdBuilder("find")
    setter(builder, "-size", "0")
    with pytest.raises(AtomicCmdBuilderError):
        builder.pop_option("-size")


def test_builder__pop_option__empty():
    builder = AtomicCmdBuilder("find")
    with pytest.raises(KeyError):
        builder.pop_option("-size")


def test_builder__pop_option__missing_key():
    builder = AtomicCmdBuilder("find")
    builder.set_option("-size", 0)
    with pytest.raises(KeyError):
        builder.pop_option("-isize")


def test_builder__pop_option__with_non_str_key():
    builder = AtomicCmdBuilder("find")
    with pytest.raises(TypeError):
        builder.pop_option(7913)


def test_builder__pop_option__with_empty_key():
    builder = AtomicCmdBuilder("find")
    with pytest.raises(KeyError):
        builder.pop_option("")


###############################################################################
###############################################################################
# AtomicCmdBuilder: add_value


def test_builder__add_value():
    builder = AtomicCmdBuilder("ls")
    builder.add_value("%(IN_FILE)s")
    assert builder.call == ["ls", "%(IN_FILE)s"]


def test_builder__add_value__two_values():
    builder = AtomicCmdBuilder("ls")
    builder.add_value("%(IN_FILE)s")
    builder.add_value("%(OUT_FILE)s")
    assert builder.call == ["ls", "%(IN_FILE)s", "%(OUT_FILE)s"]


###############################################################################
###############################################################################
# AtomicCmdBuilder: set_kwargs


def test_builder__set_kwargs__called_once():
    expected = {"IN_PATH": "/a/b/", "OUT_PATH": "/dst/file"}
    builder = AtomicCmdBuilder("echo")
    builder.set_kwargs(**expected)
    assert builder.kwargs == expected


def test_builder__set_kwargs__called_twice():
    expected = {"IN_PATH": "/a/b/", "OUT_PATH": "/dst/file"}
    builder = AtomicCmdBuilder("echo")
    builder.set_kwargs(OUT_PATH="/dst/file")
    builder.set_kwargs(IN_PATH="/a/b/")
    assert builder.kwargs == expected


def test_builder__set_kwargs__atomiccmdbuilder():
    mock = Mock(AtomicCmdBuilder("true"))
    mock.finalize.return_value = "finalized!"
    builder = AtomicCmdBuilder("ls", IN_BUILDER=mock)
    assert builder.kwargs == {"IN_BUILDER": "finalized!"}


def test_builder__set_kwargs__after_finalize():
    expected = {"IN_PATH": "/a/b/"}
    builder = AtomicCmdBuilder("echo")
    builder.set_kwargs(IN_PATH="/a/b/")
    builder.finalize()
    with pytest.raises(AtomicCmdBuilderError):
        builder.set_kwargs(OUT_PATH="/dst/file")
    assert builder.kwargs == expected


def test_builder__set__kwargs__overwriting():
    expected = {"IN_PATH": "/a/b/"}
    builder = AtomicCmdBuilder("echo")
    builder.set_kwargs(IN_PATH="/a/b/")
    with pytest.raises(AtomicCmdBuilderError):
        builder.set_kwargs(IN_PATH="/dst/file")
    assert builder.kwargs == expected


###############################################################################
###############################################################################
# AtomicCmdBuilder: finalize


def test_builder__finalized_call__simple_command():
    builder = AtomicCmdBuilder("echo")
    assert builder.finalized_call == ["echo"]


def test_builder__finalized_call__kwargs_are_instantiated():
    builder = AtomicCmdBuilder(
        ("echo", "%(ARG1)s", "X=%(ARG2)s"), ARG1="/foo/bar", ARG2="zod"
    )
    assert builder.finalized_call == ["echo", "/foo/bar", "X=zod"]


def test_builder__finalized_call__kwargs_are_instantiated__with_temp_dir():
    builder = AtomicCmdBuilder(("echo", "%(ARG)s", "%(TEMP_DIR)s"), ARG="/foo/bar")
    assert builder.finalized_call == ["echo", "/foo/bar", "%(TEMP_DIR)"]


def test_builder__finalized_call__kwargs_are_instantiated__with_non_str_arg():
    builder = AtomicCmdBuilder(("echo", "%(ARG)s", 17), ARG="/foo/bar")
    assert builder.finalized_call == ["echo", "/foo/bar", "17"]


###############################################################################
###############################################################################
# AtomicCmdBuilder: finalize


def test_builder__finalize__returns_singleton():
    builder = AtomicCmdBuilder("echo")
    assert builder.finalize() is builder.finalize()


def test_builder__finalize__calls_atomiccmd():
    builder = AtomicCmdBuilder("echo", set_cwd=True)
    builder.add_option("-out", "%(OUT_FILE)s")
    builder.add_value("%(IN_FILE)s")
    builder.set_kwargs(OUT_FILE="/out/file", IN_FILE="/in/file")
    with patch("paleomix.atomiccmd.builder.AtomicCmd") as mock:
        builder.finalize()

    assert mock.mock_calls == [
        call(
            ["echo", "-out", "%(OUT_FILE)s", "%(IN_FILE)s"],
            IN_FILE="/in/file",
            OUT_FILE="/out/file",
            set_cwd=True,
        )
    ]


###############################################################################
###############################################################################
# AtomicCmdBuilder: add_multiple_options


def test_builder__add_multiple_options():
    values = ("file_a", "file_b")
    expected = {"IN_FILE_01": "file_a", "IN_FILE_02": "file_b"}

    builder = AtomicCmdBuilder("ls")
    kwargs = builder.add_multiple_options("-i", values)

    assert kwargs == expected
    assert builder.kwargs == expected
    assert builder.call == ["ls", "-i", "%(IN_FILE_01)s", "-i", "%(IN_FILE_02)s"]


def test_builder__add_multiple_options_with_sep():
    values = ("file_a", "file_b")
    expected = {"IN_FILE_01": "file_a", "IN_FILE_02": "file_b"}

    builder = AtomicCmdBuilder("ls")
    kwargs = builder.add_multiple_options("-i", values, sep="=")

    assert kwargs == expected
    assert builder.kwargs == expected
    assert builder.call == ["ls", "-i=%(IN_FILE_01)s", "-i=%(IN_FILE_02)s"]


def test_builder__add_multiple_options_with_template():
    values = ("file_a", "file_b")
    expected = {"OUT_BAM_1": "file_a", "OUT_BAM_2": "file_b"}

    builder = AtomicCmdBuilder("ls")
    kwargs = builder.add_multiple_options("-i", values, template="OUT_BAM_%i")

    assert kwargs == expected
    assert builder.kwargs == expected
    assert builder.call == ["ls", "-i", "%(OUT_BAM_1)s", "-i", "%(OUT_BAM_2)s"]


def test_builder__add_multiple_options_multiple_times():
    expected = {"IN_FILE_01": "file_a", "IN_FILE_02": "file_b"}

    builder = AtomicCmdBuilder("ls")
    kwargs = builder.add_multiple_options("-i", ("file_a",))
    assert kwargs == {"IN_FILE_01": "file_a"}
    kwargs = builder.add_multiple_options("-i", ("file_b",))
    assert kwargs == {"IN_FILE_02": "file_b"}

    assert builder.kwargs == expected
    assert builder.call == ["ls", "-i", "%(IN_FILE_01)s", "-i", "%(IN_FILE_02)s"]


###############################################################################
###############################################################################
# AtomicCmdBuilder: add_multiple_values


def test_builder__add_multiple_values():
    values = ("file_a", "file_b")
    expected = {"IN_FILE_01": "file_a", "IN_FILE_02": "file_b"}

    builder = AtomicCmdBuilder("ls")
    kwargs = builder.add_multiple_values(values)

    assert kwargs == expected
    assert builder.kwargs == expected
    assert builder.call == ["ls", "%(IN_FILE_01)s", "%(IN_FILE_02)s"]


def test_builder__add_multiple_values_with_template():
    values = ("file_a", "file_b")
    expected = {"OUT_BAM_1": "file_a", "OUT_BAM_2": "file_b"}

    builder = AtomicCmdBuilder("ls")
    kwargs = builder.add_multiple_values(values, template="OUT_BAM_%i")

    assert kwargs == expected
    assert builder.kwargs == expected
    assert builder.call == ["ls", "%(OUT_BAM_1)s", "%(OUT_BAM_2)s"]


def test_builder__add_multiple_values_multiple_times():
    expected = {"IN_FILE_01": "file_a", "IN_FILE_02": "file_b"}

    builder = AtomicCmdBuilder("ls")
    kwargs = builder.add_multiple_values(("file_a",))
    assert kwargs == {"IN_FILE_01": "file_a"}
    kwargs = builder.add_multiple_values(("file_b",))
    assert kwargs == {"IN_FILE_02": "file_b"}

    assert builder.kwargs == expected
    assert builder.call == ["ls", "%(IN_FILE_01)s", "%(IN_FILE_02)s"]


###############################################################################
###############################################################################
# AtomicCmdBuilder: add_multiple_kwargs


def test_builder__add_multiple_kwargs():
    values = ("file_a", "file_b")
    expected = {"IN_FILE_01": "file_a", "IN_FILE_02": "file_b"}

    builder = AtomicCmdBuilder("ls")
    kwargs = builder.add_multiple_kwargs(values)

    assert kwargs == expected
    assert builder.kwargs == expected
    assert builder.call == ["ls"]


def test_builder__add_multiple_kwargs_with_template():
    values = ("file_a", "file_b")
    expected = {"OUT_BAM_1": "file_a", "OUT_BAM_2": "file_b"}

    builder = AtomicCmdBuilder("ls")
    kwargs = builder.add_multiple_kwargs(values, template="OUT_BAM_%i")

    assert kwargs == expected
    assert builder.kwargs == expected
    assert builder.call == ["ls"]


def test_builder__add_multiple_kwargs_multiple_times():
    expected = {"IN_FILE_01": "file_a", "IN_FILE_02": "file_b"}

    builder = AtomicCmdBuilder("ls")
    kwargs = builder.add_multiple_kwargs(("file_a",))
    assert kwargs == {"IN_FILE_01": "file_a"}
    kwargs = builder.add_multiple_kwargs(("file_b",))
    assert kwargs == {"IN_FILE_02": "file_b"}

    assert builder.kwargs == expected
    assert builder.call == ["ls"]


###############################################################################
###############################################################################
# AtomicJavaCmdBuilder


def test_java_builder__default__no_config():
    builder = AtomicJavaCmdBuilder("/path/Foo.jar")
    assert builder.call == [
        "java",
        "-server",
        "-Djava.io.tmpdir=%(TEMP_DIR)s",
        "-Djava.awt.headless=true",
        "-XX:+UseSerialGC",
        "-Xmx4g",
        "-jar",
        "%(AUX_JAR)s",
    ]


def test_java_builder__defaults__call():
    builder = AtomicJavaCmdBuilder("/path/Foo.jar", temp_root="/disk/tmp")
    assert builder.call == [
        "java",
        "-server",
        "-Djava.io.tmpdir=/disk/tmp",
        "-Djava.awt.headless=true",
        "-XX:+UseSerialGC",
        "-Xmx4g",
        "-jar",
        "%(AUX_JAR)s",
    ]


def test_java_builder__defaults__kwargs():
    builder = AtomicJavaCmdBuilder("/path/Foo.jar")
    assert builder.kwargs == {"AUX_JAR": "/path/Foo.jar"}


def test_java_builder__multithreaded_gc():
    builder = AtomicJavaCmdBuilder("/path/Foo.jar", temp_root="/disk/tmp", gc_threads=3)
    assert builder.call == [
        "java",
        "-server",
        "-Djava.io.tmpdir=/disk/tmp",
        "-Djava.awt.headless=true",
        "-XX:ParallelGCThreads=3",
        "-Xmx4g",
        "-jar",
        "%(AUX_JAR)s",
    ]


def test_java_builder__multithreaded_gc__zero_or_negative_threads():
    with pytest.raises(ValueError):
        AtomicJavaCmdBuilder("/path/Foo.jar", gc_threads=0)
    with pytest.raises(ValueError):
        AtomicJavaCmdBuilder("/path/Foo.jar", gc_threads=-1)


def test_java_builder__multithreaded_gc__non_int_threads():
    with pytest.raises(TypeError):
        AtomicJavaCmdBuilder("/path/Foo.jar", gc_threads="3")


def test_java_builder__kwargs():
    builder = AtomicJavaCmdBuilder("/path/Foo.jar", set_cwd=True)
    assert builder.kwargs == {"AUX_JAR": "/path/Foo.jar", "set_cwd": True}


def test_java_builder__default__jre_options():
    builder = AtomicJavaCmdBuilder("/path/Foo.jar", jre_options=("-XtraOption",))
    assert builder.call == [
        "java",
        "-server",
        "-Djava.io.tmpdir=%(TEMP_DIR)s",
        "-Djava.awt.headless=true",
        "-XX:+UseSerialGC",
        "-XtraOption",
        "-Xmx4g",
        "-jar",
        "%(AUX_JAR)s",
    ]


def test_java_builder__default__jre_options__Xmx():
    builder = AtomicJavaCmdBuilder("/path/Foo.jar", jre_options=("-Xmx8g",))
    assert builder.call == [
        "java",
        "-server",
        "-Djava.io.tmpdir=%(TEMP_DIR)s",
        "-Djava.awt.headless=true",
        "-XX:+UseSerialGC",
        "-Xmx8g",
        "-jar",
        "%(AUX_JAR)s",
    ]


###############################################################################
###############################################################################
# AtomicMPICmdBuilder


def test_mpi_builder__defaults__str():
    builder = AtomicMPICmdBuilder("ls")
    assert builder.call == ["ls"]
    assert builder.kwargs == {"EXEC_MPI": "mpirun"}


def test_mpi_builder__multithreaded__str():
    builder = AtomicMPICmdBuilder("ls", threads=3)
    assert builder.call == ["mpirun", "-n", 3, "ls"]
    assert builder.kwargs == {"EXEC_MAIN": "ls"}


def test_mpi_builder__defaults__complex_cmd():
    builder = AtomicMPICmdBuilder(["python", "/foo/run.py"])
    assert builder.call == ["python", "/foo/run.py"]
    assert builder.kwargs == {"EXEC_MPI": "mpirun"}


def test_mpi_builder__multithreaded__complex_cmd():
    builder = AtomicMPICmdBuilder(["python", "/foo/run.py"], threads=3)
    assert builder.call == ["mpirun", "-n", 3, "python", "/foo/run.py"]
    assert builder.kwargs == {"EXEC_MAIN": "python"}


def test_mpi_builder__kwargs():
    builder = AtomicMPICmdBuilder("ls", set_cwd=True)
    assert builder.kwargs == {"set_cwd": True, "EXEC_MPI": "mpirun"}


def test_mpi_builder__threads__zero_or_negative():
    with pytest.raises(ValueError):
        AtomicMPICmdBuilder("ls", threads=0)
    with pytest.raises(ValueError):
        AtomicMPICmdBuilder("ls", threads=-1)


def test_mpi_builder__threads__non_int():
    with pytest.raises(TypeError):
        AtomicMPICmdBuilder("ls", threads="3")


###############################################################################
###############################################################################
# apply_options


def test_apply_options__single_option__default_pred__set_when_pred_is_true():
    mock = Mock(AtomicCmdBuilder("ls"))
    apply_options(mock, {"--foo": 17})
    mock.set_option.assert_called_once_with("--foo", 17)


def test_apply_options__single_option__default_pred__ignore_false_pred():
    mock = Mock(AtomicCmdBuilder("ls"))
    apply_options(mock, {"Other": None})


def _user_pred(key):
    return key.startswith("FOO")


def test_apply_options__single_option__user_pred__set_when_pred_is_true():
    mock = Mock(AtomicCmdBuilder("ls"))
    apply_options(mock, {"FOO_BAR": 17}, _user_pred)
    mock.set_option.assert_called_once_with("FOO_BAR", 17)


def test_apply_options__single_option__user_pred__ignore_when_pred_is_false():
    mock = Mock(AtomicCmdBuilder("ls"))
    apply_options(mock, {"BAR_FOO": 17}, _user_pred)


def test_apply_options__single_option__boolean__set_when_value_is_true():
    mock = Mock(AtomicCmdBuilder("ls"))
    apply_options(mock, {"-v": True})
    mock.set_option.assert_called_once_with("-v")


def test_apply_options__single_option__boolean__set_when_value_is_none():
    mock = Mock(AtomicCmdBuilder("ls"))
    apply_options(mock, {"-v": None})
    mock.set_option.assert_called_once_with("-v")


def test_apply_options__single_option__boolean__pop_when_value_is_false():
    mock = Mock(AtomicCmdBuilder("ls"))
    apply_options(mock, {"-v": False})
    mock.pop_option.assert_called_once_with("-v")


def test_apply_options__single_option__boolean__pop_missing_throws():
    mock = Mock(AtomicCmdBuilder("ls"))
    mock.pop_option.side_effect = KeyError("-v")
    with pytest.raises(KeyError):
        apply_options(mock, {"-v": False})
    mock.pop_option.assert_called_once_with("-v")


def test_apply_options__multiple_option():
    mock = Mock(AtomicCmdBuilder("ls"))
    apply_options(mock, {"--foo": [3, 17]})
    assert mock.mock_calls == [
        call.add_option("--foo", 3),
        call.add_option("--foo", 17),
    ]


def test_apply_options__boolean_and_none_is_single_value_only():
    mock = Mock(AtomicCmdBuilder("ls"))
    with pytest.raises(TypeError):
        apply_options(mock, {"--foo": [True]})
    with pytest.raises(TypeError):
        apply_options(mock, {"--foo": [False]})
    with pytest.raises(TypeError):
        apply_options(mock, {"--foo": [None]})


def test_apply_options__unexpected_types_in_values():
    mock = Mock(AtomicCmdBuilder("ls"))
    with pytest.raises(TypeError):
        apply_options(mock, {"--foo": object()})
    with pytest.raises(TypeError):
        apply_options(mock, {"--foo": iter([])})
    with pytest.raises(TypeError):
        apply_options(mock, {"--foo": {}})
    with pytest.raises(TypeError):
        apply_options(mock, {"--foo": set()})


def test_apply_options__non_string_types_in_keys():
    mock = Mock(AtomicCmdBuilder("ls"))
    with pytest.raises(TypeError):
        apply_options(mock, {1: 17})
    with pytest.raises(TypeError):
        apply_options(mock, {("foo",): 17})


def test_apply_options__not_dict_like():
    mock = Mock(AtomicCmdBuilder("ls"))
    with pytest.raises(TypeError):
        apply_options(mock, None)
    with pytest.raises(TypeError):
        apply_options(mock, [1, 2, 3])



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import errno
import logging
import multiprocessing
import os
import signal
import sys
import traceback
from queue import Empty

import paleomix.common.logging
from paleomix.common.text import padded_table
from paleomix.common.utilities import safe_coerce_to_tuple
from paleomix.common.versions import VersionRequirementError
from paleomix.node import Node, NodeError, NodeUnhandledException
from paleomix.nodegraph import FileStatusCache, NodeGraph, NodeGraphError


class Pypeline:
    def __init__(self, config):
        self._nodes = []
        self._config = config
        self._logger = logging.getLogger(__name__)
        # Set if a keyboard-interrupt (SIGINT) has been caught
        self._interrupted = False
        self._queue = multiprocessing.Queue()
        self._pool = None

    def add_nodes(self, *nodes):
        for subnodes in safe_coerce_to_tuple(nodes):
            for node in safe_coerce_to_tuple(subnodes):
                if not isinstance(node, Node):
                    raise TypeError("Node object expected, received %s" % repr(node))
                self._nodes.append(node)

    def run(self, max_threads=1, dry_run=False):
        if max_threads < 1:
            raise ValueError("Max threads must be >= 1")

        try:
            nodegraph = NodeGraph(self._nodes)
        except NodeGraphError as error:
            self._logger.error(error)
            return False

        for node in nodegraph.iterflat():
            if node.threads > max_threads:
                self._logger.warn(
                    "One or more tasks require more threads than the user-defined "
                    "maximum; the number of threads used will therefore exceed the "
                    "user-defined maximum when running those tasks."
                )
                break

        if dry_run:
            self._summarize_pipeline(nodegraph)
            self._logger.info("Dry run done")

            result = True
        else:
            self._pool = multiprocessing.Pool(max_threads, _init_worker, (self._queue,))
            old_handler = signal.signal(signal.SIGINT, self._sigint_handler)

            try:
                result = self._run(nodegraph, max_threads)
            finally:
                signal.signal(signal.SIGINT, old_handler)

        for filename in paleomix.common.logging.get_logfiles():
            self._logger.info("Log-file written to %r", filename)

        return result

    def _run(self, nodegraph, max_threads):
        # Dictionary of nodes -> async-results
        running = {}
        # Set of remaining nodes to be run
        remaining = set(nodegraph.iterflat())

        is_ok = True
        while running or (remaining and not self._interrupted):
            is_ok &= self._poll_running_nodes(running, nodegraph, self._queue)

            if not self._interrupted:  # Prevent starting of new nodes
                self._start_new_tasks(
                    remaining, running, nodegraph, max_threads, self._pool
                )

        self._pool.close()
        self._pool.join()
        self._summarize_pipeline(nodegraph)

        return is_ok

    def _start_new_tasks(self, remaining, running, nodegraph, max_threads, pool):
        started_nodes = []
        idle_processes = max_threads - sum(
            node.threads for (node, _) in running.values()
        )

        if not idle_processes:
            return False

        for node in remaining:
            if not running or (idle_processes >= node.threads):
                state = nodegraph.get_node_state(node)
                if state == nodegraph.RUNABLE:
                    key = id(node)
                    proc_args = (key, node, self._config)
                    running[key] = (node, pool.apply_async(_call_run, args=proc_args))
                    started_nodes.append(node)

                    nodegraph.set_node_state(node, nodegraph.RUNNING)
                    idle_processes -= node.threads
                elif state in (nodegraph.DONE, nodegraph.ERROR):
                    started_nodes.append(node)
            elif idle_processes <= 0:
                break

        for node in started_nodes:
            remaining.remove(node)

    def _poll_running_nodes(self, running, nodegraph, queue):
        error_happened = False
        blocking = False

        while running and not error_happened:
            node, proc = self._get_finished_node(queue, running, blocking)
            if not node:
                if blocking:
                    break

                blocking = True
                continue

            try:
                # Re-raise exceptions from the node-process
                proc.get()
            except (KeyboardInterrupt, SystemExit):
                raise
            except Exception as errors:
                error_happened = True
                nodegraph.set_node_state(node, nodegraph.ERROR)

                self._logger.error("%s while %s:", type(errors).__name__, node)
                for line in str(errors).strip().split("\n"):
                    self._logger.error("    %s", line)

            if not error_happened:
                nodegraph.set_node_state(node, nodegraph.DONE)

        return not error_happened

    @property
    def nodes(self):
        return set(self._nodes)

    def walk_nodes(self, func):
        skip_nodes = set()

        def _walk_nodes(nodes):
            for node in nodes:
                if node in skip_nodes:
                    continue
                elif not func(node):
                    return False

                skip_nodes.add(node)
                if not _walk_nodes(node.dependencies):
                    return False
            return True

        _walk_nodes(self._nodes)

    def list_input_files(self):
        """Returns a set containing the absolute path of all input files
        required by the current pipeline. These do not include any file
        generated by the pipeline itself (output files).
        """
        input_files = set()
        output_files = set()

        def collect_output_files(node):
            for filename in node.input_files:
                input_files.add(os.path.abspath(filename))

            for filename in node.output_files:
                output_files.add(os.path.abspath(filename))

            return True

        self.walk_nodes(collect_output_files)

        return input_files - output_files

    def list_output_files(self):
        cache = FileStatusCache()
        nodegraph = NodeGraph(
            self._nodes,
            allow_missing_files=True,
            cache_factory=lambda: cache,
        )
        output_files = {}

        def collect_output_files(node):
            state = None
            if nodegraph.is_done(node, cache):
                state = nodegraph.DONE
                if nodegraph.is_outdated(node, cache):
                    state = nodegraph.OUTDATED

            for filename in node.output_files:
                output_files[os.path.abspath(filename)] = state

            return True

        self.walk_nodes(collect_output_files)

        return output_files

    def list_required_executables(self):
        requirements = {}

        def collect_requirements(node):
            for executable in node.executables:
                if executable not in requirements:
                    requirements[executable] = set()

            for requirement in node.requirements:
                if requirement.name not in requirements:
                    requirements[requirement.name] = set()

                requirements[requirement.name].add(requirement)

                executable = requirement.executable
                if not requirements.get(executable):
                    requirements.pop(executable, None)

            return True

        self.walk_nodes(collect_requirements)
        return requirements

    def print_output_files(self, print_func=print):
        output_files = self.list_output_files()

        for filename, state in sorted(output_files.items()):
            if state == NodeGraph.DONE:
                state = "Ready      "
            elif state == NodeGraph.OUTDATED:
                state = "Outdated   "
            else:
                state = "Missing    "

            print_func("%s\t%s" % (state, filename))

    def print_input_files(self, print_func=print):
        """Prints the absolute path of all input files required by the current
        pipeline, excluding any file generated by the pipeline itself (output
        files). One file is printed per line.
        """
        input_files = self.list_input_files()
        for filename in sorted(input_files):
            print_func("%s" % (filename,))

    def print_required_executables(self, print_func=print):
        template = "{: <40s} {: <11s} {}"
        pipeline_executables = self.list_required_executables()
        print_func(template.format("Executable", "Version", "Required version"))

        for name, requirements in sorted(pipeline_executables.items()):
            if not requirements:
                print_func(template.format(name, "-", "any version"))
                continue

            for requirement in requirements:
                try:
                    if requirement.version:
                        version = "v" + ".".join(map(str, requirement.version))
                    else:
                        version = "NA"
                except VersionRequirementError:
                    version = "UNKNOWN"

                print_func(template.format(name, version, requirement.checks))

    def _sigint_handler(self, signum, frame):
        """Signal handler; see signal.signal."""
        if not self._interrupted:
            self._interrupted = True
            self._logger.warning(
                "Keyboard interrupt detected, waiting for running tasks to complete. "
                "Press CTRL-C again to force termination."
            )
        else:
            self._logger.warning("Terminating pipeline!")
            self._pool.terminate()
            sys.exit(-signum)

    @classmethod
    def _get_finished_node(cls, queue, running, blocking):
        """Returns a tuple containing a node that has finished running
        and it's async-result, or None for both if no such node could
        be found (and blocking is False), or if an interrupt occurred
        while waiting for a node to finish.

        If blocking is True, the function will timeout after 0.1s.
        """
        try:
            key = queue.get(blocking, 0.1)
            return running.pop(key)
        except IOError as error:
            # User pressed ctrl-c (SIGINT), or similar event
            if error.errno != errno.EINTR:
                raise
        except Empty:
            pass
        return None, None

    def _summarize_pipeline(self, nodegraph):
        states = nodegraph.get_state_counts()

        rows = [
            ("Number of nodes:", sum(states)),
            ("Number of done nodes:", states[nodegraph.DONE]),
            ("Number of runable nodes:", states[nodegraph.RUNABLE]),
            ("Number of queued nodes:", states[nodegraph.QUEUED]),
            ("Number of outdated nodes:", states[nodegraph.OUTDATED]),
            ("Number of failed nodes:", states[nodegraph.ERROR]),
        ]

        for message in padded_table(rows):
            self._logger.info(message)

        if states[nodegraph.ERROR]:
            self._logger.warning("Errors were detected while running pipeline")
        else:
            self._logger.info("Pipeline completed successfully")


def _init_worker(queue):
    """Init function for subprocesses created by multiprocessing.Pool: Ensures
    that KeyboardInterrupts only occur in the main process, allowing us to do
    proper cleanup.
    """
    signal.signal(signal.SIGINT, signal.SIG_IGN)
    # This is a workaround to avoid having to use multiprocessing.Manager
    # to create the Queue objects; this is needed because the Manager class
    # creates it own process, which inherits the signal-handlers of the main
    # process, causing some rather odd behavior when the user causes a SIGINT.
    _call_run.queue = queue


def _call_run(key, node, config):
    """Wrapper function, required in order to call Node.run()
    in subprocesses, since it is not possible to pickle
    bound functions (e.g. self.run)"""
    try:
        return node.run(config)
    except NodeError:
        raise
    except Exception:
        message = "Unhandled error running Node:\n\n%s" % (traceback.format_exc(),)

        raise NodeUnhandledException(message)
    finally:
        # See comment in _init_worker
        _call_run.queue.put(key)



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import bz2
import copy
import gzip
import io
from unittest.mock import patch

import pytest

from paleomix.common.fileutils import fspath
from paleomix.common.formats.fasta import FASTA
from paleomix.common.formats.msa import MSA, FASTAError, MSAError

###############################################################################
###############################################################################
# Tests for constructor


def test_msa_constructor__calls_validate():
    with patch("paleomix.common.formats.msa.MSA.validate", wrap=MSA.validate) as mock:
        MSA([FASTA("NA", None, "ACGT")])

    assert len(mock.mock_calls) == 1


def test_msa_constructor__duplicate_names():
    records = [FASTA("Foo", None, "ACGT"), FASTA("Foo", None, "GTCA")]
    with pytest.raises(MSAError):
        MSA(records)


def test_msa_constructor__empty_msa():
    with pytest.raises(MSAError):
        MSA([])


###############################################################################
###############################################################################
# Tests for 'seqlen' / len


def test_msa__len__corresponds_to_sequence_number_of_records():
    msa = MSA(
        (
            FASTA("seq1", None, "ACGCGTATGCATGCCGA"),
            FASTA("seq2", None, "TGAACACACAGTAGGAT"),
        )
    )
    assert len(msa) == 2


def test_msa__seqlen__corresponds_to_sequence_lengths():
    msa = MSA(
        (
            FASTA("seq1", None, "ACGCGTATGCATGCCGA"),
            FASTA("seq2", None, "TGAACACACAGTAGGAT"),
        )
    )
    assert msa.seqlen() == 17


###############################################################################
###############################################################################
# Tests for 'exclude'


def test_msa_exclude__remove_one():
    fa_1 = FASTA("A", None, "ACGT")
    fa_2 = FASTA("B", None, "GCTA")
    initial = MSA([fa_1, fa_2])
    expected = MSA([fa_1])
    result = initial.exclude(["B"])
    assert result == expected


def test_msa_exclude__missing_keys():
    msa = MSA([FASTA("Foo", None, "ACGT")])
    with pytest.raises(KeyError):
        msa.exclude(["Bar"])


def test_msa_exclude__no_keys():
    msa = MSA([FASTA("Foo", None, "ACGT")])
    with pytest.raises(ValueError):
        msa.exclude([])


###############################################################################
###############################################################################
# Tests for 'select'


def test_msa_select__remove_one():
    fa_1 = FASTA("A", None, "ACGT")
    fa_2 = FASTA("B", None, "GCTA")
    initial = MSA([fa_1, fa_2])
    expected = MSA([fa_1])
    result = initial.select(["A"])
    assert result == expected


def test_msa_select__missing_keys():
    msa = MSA([FASTA("Foo", None, "ACGT")])
    with pytest.raises(KeyError):
        msa.select(["Bar"])


def test_msa_select__no_keys():
    msa = MSA([FASTA("Foo", None, "ACGT")])
    with pytest.raises(ValueError):
        msa.select([])


###############################################################################
###############################################################################
# Tests for 'reduce'


def test_msa_reduce__no_empty_columns__no_columns_are_removed():
    fa_1 = FASTA("Name_A", "Meta_A", "ACnT")
    fa_2 = FASTA("Name_B", "Meta_B", "C-TN")
    initial = MSA([fa_1, fa_2])
    expected = MSA([fa_1, fa_2])
    assert initial.reduce() == expected


def test_msa_reduce__one_empty_column__column_are_removed():
    fa_1 = FASTA("Name_A", "Meta_A", "AnT")
    fa_2 = FASTA("Name_B", "Meta_B", "C-N")
    initial = MSA([fa_1, fa_2])
    fa_reduced_1 = FASTA("Name_A", "Meta_A", "AT")
    fa_reduced_2 = FASTA("Name_B", "Meta_B", "CN")
    expected = MSA([fa_reduced_1, fa_reduced_2])
    assert initial.reduce() == expected


def test_msa_reduce__multiple_empty_column__all_empty_column_are_removed():
    fa_1 = FASTA("Name_A", "Meta_A", "-AnTN")
    fa_2 = FASTA("Name_B", "Meta_B", "NC-NN")
    initial = MSA([fa_1, fa_2])
    fa_reduced_1 = FASTA("Name_A", "Meta_A", "AT")
    fa_reduced_2 = FASTA("Name_B", "Meta_B", "CN")
    expected = MSA([fa_reduced_1, fa_reduced_2])
    assert initial.reduce() == expected


def test_msa_reduce__only_empty_column__none_is_returned():
    fa_1 = FASTA("Name_A", "Meta_A", "---Nn")
    fa_2 = FASTA("Name_B", "Meta_B", "Nn--N")
    initial = MSA([fa_1, fa_2])
    assert initial.reduce() is None


###############################################################################
###############################################################################
# Tests for 'filter_singletons'

_FILTER_MSA_1 = MSA(
    (
        FASTA("Seq1", "Meta1", "ACGNTYCSTG"),
        FASTA("Seq2", "Meta2", "ACTA-WCCTG"),
        FASTA("Seq3", "Meta3", "NCGGTYCGTC"),
    )
)


def test_msa_filter_singletons__filter_by_second():
    expected = MSA(
        (
            FASTA("Seq1", "Meta1", "ACnNntCcTG"),
            FASTA("Seq2", "Meta2", "ACTA-WCCTG"),
            FASTA("Seq3", "Meta3", "NCGGTYCGTC"),
        )
    )
    result = _FILTER_MSA_1.filter_singletons("Seq1", ["Seq2"])
    assert result == expected


def test_msa_filter_singletons__filter_by_third():
    expected = MSA(
        (
            FASTA("Seq1", "Meta1", "nCGNTYCgTn"),
            FASTA("Seq2", "Meta2", "ACTA-WCCTG"),
            FASTA("Seq3", "Meta3", "NCGGTYCGTC"),
        )
    )
    result = _FILTER_MSA_1.filter_singletons("Seq1", ["Seq3"])
    assert result == expected


def test_msa_filter_singletons__filter_by_both():
    result = _FILTER_MSA_1.filter_singletons("Seq1", ["Seq2", "Seq3"])
    assert result == _FILTER_MSA_1


def test_msa_filter_singletons__filter_by_itself():
    with pytest.raises(MSAError):
        _FILTER_MSA_1.filter_singletons("Seq1", ["Seq1", "Seq2"])


def test_msa_filter_singletons__filter_by_nothing():
    with pytest.raises(ValueError):
        _FILTER_MSA_1.filter_singletons("Seq1", [])


###############################################################################
###############################################################################
# Tests for 'MSA.join'

_JOIN_MSA_1 = MSA(
    (FASTA("nc", None, "ACG"), FASTA("nm", None, "TGA"), FASTA("miRNA", None, "UCA"))
)
_JOIN_MSA_2 = MSA(
    (FASTA("nc", None, "TGA"), FASTA("nm", None, "CTT"), FASTA("miRNA", None, "GAC"))
)
_JOIN_MSA_3 = MSA(
    (FASTA("nc", None, "AAG"), FASTA("nm", None, "GAG"), FASTA("miRNA", None, "CAU"))
)


def test_msa_join__single_msa():
    result = MSA.join(_JOIN_MSA_1)
    assert result == _JOIN_MSA_1


def test_msa_join__two_msa():
    expected = MSA(
        (
            FASTA("nc", None, "ACGTGA"),
            FASTA("nm", None, "TGACTT"),
            FASTA("miRNA", None, "UCAGAC"),
        )
    )
    result = MSA.join(_JOIN_MSA_1, _JOIN_MSA_2)
    assert result == expected


def test_msa_join__three_msa():
    expected = MSA(
        (
            FASTA("nc", None, "ACGTGAAAG"),
            FASTA("nm", None, "TGACTTGAG"),
            FASTA("miRNA", None, "UCAGACCAU"),
        )
    )
    result = MSA.join(_JOIN_MSA_1, _JOIN_MSA_2, _JOIN_MSA_3)
    assert result == expected


def test_msa_join__missing_arguments():
    with pytest.raises(TypeError):
        MSA.join()


###############################################################################
###############################################################################
# Tests for 'MSA.from_lines'


def test_msa_from_lines__single_entry():
    lines = [">seq1", "ACG"]
    result = MSA([FASTA("seq1", None, "ACG")])
    assert MSA.from_lines(lines) == result


def test_msa_from_lines__single_entry_with_meta():
    lines = [">seq1 Meta info", "ACG"]
    expected = MSA([FASTA("seq1", "Meta info", "ACG")])
    result = MSA.from_lines(lines)
    assert result == expected


def test_msa_from_lines__two_entries():
    lines = [">seq1", "ACG", ">seq2", "TGA"]
    expected = MSA([FASTA("seq1", None, "ACG"), FASTA("seq2", None, "TGA")])
    result = MSA.from_lines(lines)
    assert result == expected


def test_msa_from_lines__two_entries_with_meta():
    lines = [">seq1", "ACG", ">seq2 Second meta", "TGA"]
    expected = MSA([FASTA("seq1", None, "ACG"), FASTA("seq2", "Second meta", "TGA")])
    result = MSA.from_lines(lines)
    assert result == expected


def test_msa_from_lines__duplicate_names():
    with pytest.raises(MSAError):
        MSA.from_lines([">seq1", "ACG", ">seq1", "TGA"])


def test_msa_from_lines__mismatched_lengths():
    with pytest.raises(MSAError):
        MSA.from_lines([">seq1", "ACG", ">seq2", "TGAN"])


def test_msa_from_lines__empty_name():
    with pytest.raises(FASTAError):
        MSA.from_lines([">", "ACG", ">seq1", "TGAN"])


###############################################################################
###############################################################################
# Tests for 'MSA.from_file'


@pytest.mark.parametrize("func", (open, gzip.open, bz2.open))
def test_msa_from_file(func, tmp_path):
    filename = tmp_path / "test.fasta"
    with func(fspath(filename), "wt") as handle:
        handle.write(">This_is_FASTA!\nACGTN\n>This_is_ALSO_FASTA!\nCGTNA\n")

    assert MSA.from_file(filename) == MSA(
        [
            FASTA("This_is_FASTA!", None, "ACGTN"),
            FASTA("This_is_ALSO_FASTA!", None, "CGTNA"),
        ]
    )


###############################################################################
###############################################################################
# Tests for 'MSA.split'


def test_msa_split_msa__single_group():
    msa = MSA([FASTA("seq1", None, "ACGCAT"), FASTA("seq2", None, "GAGTGA")])
    expected = {"1": copy.copy(msa)}
    assert msa.split("111") == expected


def test_msa_split_msa__two_groups():
    msa = MSA([FASTA("seq1", None, "ACGCAT"), FASTA("seq2", None, "GAGTGA")])
    expected = {
        "1": MSA([FASTA("seq1", None, "ACCA"), FASTA("seq2", None, "GATG")]),
        "2": MSA([FASTA("seq1", None, "GT"), FASTA("seq2", None, "GA")]),
    }
    assert msa.split("112") == expected


def test_msa_split__three_groups():
    msa = MSA([FASTA("seq1", None, "ACGCAT"), FASTA("seq2", None, "GAGTGA")])
    expected = {
        "1": MSA([FASTA("seq1", None, "AC"), FASTA("seq2", None, "GT")]),
        "2": MSA([FASTA("seq1", None, "CA"), FASTA("seq2", None, "AG")]),
        "3": MSA([FASTA("seq1", None, "GT"), FASTA("seq2", None, "GA")]),
    }
    assert msa.split("123") == expected


def test_msa_split__empty_group():
    msa = MSA([FASTA("seq1", None, "AC"), FASTA("seq2", None, "GA")])
    expected = {
        "1": MSA([FASTA("seq1", None, "A"), FASTA("seq2", None, "G")]),
        "2": MSA([FASTA("seq1", None, "C"), FASTA("seq2", None, "A")]),
        "3": MSA([FASTA("seq1", None, ""), FASTA("seq2", None, "")]),
    }
    assert msa.split("123") == expected


def test_msa_split__partial_group():
    msa = MSA([FASTA("seq1", None, "ACGCA"), FASTA("seq2", None, "GAGTG")])
    expected = {
        "1": MSA([FASTA("seq1", None, "AC"), FASTA("seq2", None, "GT")]),
        "2": MSA([FASTA("seq1", None, "CA"), FASTA("seq2", None, "AG")]),
        "3": MSA([FASTA("seq1", None, "G"), FASTA("seq2", None, "G")]),
    }
    assert msa.split("123") == expected


def test_msa_split_msa__no_split_by():
    msa = MSA([FASTA("seq1", None, "ACG"), FASTA("seq2", None, "GAT")])
    with pytest.raises(TypeError):
        msa.split(split_by="")


###############################################################################
###############################################################################
# Tests for 'MSA.to_file'


def test_msa_to_file__complete_line_test():
    msa = MSA(
        [
            FASTA("barfoo", None, "ACGATA" * 10 + "CGATAG" * 5),
            FASTA("foobar", None, "CGAATG" * 10 + "TGTCAT" * 5),
        ]
    )
    expected = ">barfoo\n%s\n%s\n" % ("ACGATA" * 10, "CGATAG" * 5)
    expected += ">foobar\n%s\n%s\n" % ("CGAATG" * 10, "TGTCAT" * 5)
    stringf = io.StringIO()
    MSA.to_file(msa, stringf)
    assert stringf.getvalue() == expected


###############################################################################
###############################################################################
# Tests for 'MSA.validate'


def test_msa_validate__missing_names_first():
    msa_1 = MSA(list(_JOIN_MSA_1)[:-1])
    msa_2 = copy.copy(_JOIN_MSA_2)
    with pytest.raises(MSAError):
        MSA.validate(msa_1, msa_2)


def test_msa_validate__missing_names_second():
    msa_1 = copy.copy(_JOIN_MSA_1)
    msa_2 = MSA(list(_JOIN_MSA_2)[:-1])
    with pytest.raises(MSAError):
        MSA.validate(msa_1, msa_2)


###############################################################################
###############################################################################
# Tests for 'MSA.names'


def test_msa_names():
    assert _JOIN_MSA_1.names() == set(("nc", "nm", "miRNA"))


###############################################################################
###############################################################################
# Tests for str/repr


def test_msa_repr():
    msa = MSA(
        (
            FASTA("nc", None, "ACGTA"),
            FASTA("nm", "META", "TGAGT"),
            FASTA("miRNA", None, "UCAGA"),
        )
    )

    assert (
        str(msa) == "MSA(FASTA('miRNA', '', 'UCAGA'), "
        "FASTA('nc', '', 'ACGTA'), "
        "FASTA('nm', 'META', 'TGAGT'))"
    )


def test_msa_repr__same_as_str():
    assert str(_JOIN_MSA_1) == repr(_JOIN_MSA_1)



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import collections
import os
import re

import pysam

from paleomix.common.fileutils import describe_files, make_dirs
from paleomix.common.sequences import reverse_complement
from paleomix.common.utilities import chain_sorted
from paleomix.node import CommandNode, Node, NodeError
from paleomix.tools import factory


class DetectInputDuplicationNode(Node):
    """Attempts to detect reads included multiple times as input based on the
    presence of reads with identical names AND sequences. This is compromise
    between sensitivity, specificity, and running time.

    A possible refinement would be to consider reads with the same name where
    one read is the prefix of the other (due to different amounts of trimming
    or collapsing of reads).
    """

    def __init__(self, input_files, output_file, dependencies=()):
        Node.__init__(
            self,
            description="detecting duplicate input in %s"
            % (describe_files(input_files)),
            input_files=input_files,
            output_files=output_file,
            dependencies=dependencies,
        )

    def run(self, _):
        check_bam_files(self.input_files, self._throw_node_error)

        # Everything is ok, touch the output files
        for fpath in self.output_files:
            if os.path.dirname(fpath):
                make_dirs(os.path.dirname(fpath))

            with open(fpath, "w"):
                pass

    def _throw_node_error(self, chrom, pos, records, name, seq, qual):
        message = [
            "The same read was found multiple times at position %i on %r!"
            % (pos, chrom),
            "    Name:      %r" % (name,),
            "    Sequence:  %r" % (seq,),
            "    Qualities: %r" % (qual,),
            "",
        ]

        message.append("Read was found in these BAM files:")
        for filename, records in sorted(records.items()):
            message.append("   %s in %r" % (_summarize_reads(records), filename))

        message.append("")
        message.append(
            "This indicates that the same data has been "
            "included multiple times in the project. This "
            "can be because multiple copies of the same "
            "files were used, or because one or more files "
            "contain multiple copies of the same reads. "
            "The command 'paleomix dupcheck' may be used "
            "to review the potentially duplicated data in "
            "these BAM files.\n\n"
            "If this error was a false positive, then you "
            "may execute the following command(s) to mark "
            "this test as having succeeded:"
        )

        for fpath in self.output_files:
            message.append("$ touch '%s'" % (fpath,))

        raise NodeError("\n".join(message))


class ValidateFASTQFilesNode(CommandNode):
    def __init__(
        self, input_files, output_file, offset, collapsed=False, dependencies=()
    ):
        command = factory.new(":validate_fastq")
        command.set_option("--offset", offset)
        if collapsed:
            command.set_option("--collapsed")
        command.add_multiple_values(input_files)
        command.set_kwargs(OUT_STDOUT=output_file)

        CommandNode.__init__(
            self,
            description="validating %s" % (describe_files(input_files),),
            command=command.finalize(),
            dependencies=dependencies,
        )


class ValidateFASTAFilesNode(Node):
    def __init__(self, input_files, output_file, dependencies=()):
        Node.__init__(
            self,
            description="validating %s" % (describe_files(input_files),),
            input_files=input_files,
            output_files=output_file,
            dependencies=dependencies,
        )

        assert len(self.output_files) == 1, self.output_files

    def _run(self, _config, _temp):
        for filename in self.input_files:
            check_fasta_file(filename)
        (output_file,) = self.output_files
        if os.path.dirname(output_file):
            make_dirs(os.path.dirname(output_file))
        with open(output_file, "w"):
            pass


def check_bam_files(input_files, err_func):
    handles = []
    try:
        last_pos = None
        observed_reads = collections.defaultdict(list)
        reads_iter = _open_samfiles(handles, input_files)
        references = handles[0].references

        for record, filename in reads_iter:
            curr_pos = (record.pos, record.tid)
            if curr_pos != last_pos:
                _process_bam_reads(observed_reads, references, last_pos, err_func)
                observed_reads.clear()
                last_pos = curr_pos

                # Stop once the trailing, unmapped reads are reached
                if record.tid == -1:
                    break

            observed_reads[record.qname].append((record, filename))
        _process_bam_reads(observed_reads, references, last_pos, err_func)
    finally:
        for handle in handles:
            handle.close()


def _open_samfiles(handles, filenames):
    sequences = []
    for filename in filenames:
        handle = pysam.AlignmentFile(filename)
        handles.append(handle)

        sequences.append(_read_samfile(handle, filename))

    return chain_sorted(*sequences, key=_key_by_tid_pos)


def _read_samfile(handle, filename):
    for record in handle:
        if record.is_unmapped and (not record.pos or record.mate_is_unmapped):
            # Ignore unmapped reads except when these are sorted
            # according to the mate position (if mapped)
            continue
        elif record.flag & 0x900:
            # Ignore supplementary / secondary alignments
            continue

        yield (record, filename)


def _process_bam_reads(observed_reads, references, position, err_func):
    for records_and_filenames in observed_reads.values():
        if len(records_and_filenames) == 1:
            # Most read-names should be obseved at most once at a position
            continue

        result = collections.defaultdict(list)
        for record, filename in records_and_filenames:
            key = (record.is_reverse, record.qname, record.seq, record.qual)
            result[key].append((filename, record))

        for (is_reverse, name, seq, qual), filenames in result.items():
            if len(filenames) == 1:
                # Two reads had same name, but different characteristics
                continue

            records = collections.defaultdict(list)
            for filename, record in filenames:
                records[filename].append(record)

            if is_reverse:
                seq = reverse_complement(seq)
                qual = qual[::-1]

            chrom = references[position[1]]
            pos = position[0]

            err_func(chrom, pos, records, name, seq, qual)


def _summarize_reads(records):
    counts = {"mate 1": 0, "mate 2": 0, "unpaired": 0}

    for record in records:
        if record.is_paired:
            if record.is_read1:
                counts["mate 1"] += 1
            elif record.is_read2:
                counts["mate 2"] += 1
            else:
                counts["unpaired"] += 1
        else:
            counts["unpaired"] += 1

    result = []
    for key, value in sorted(counts.items()):
        if value > 1:
            result.append("%i %s reads" % (value, key))
        elif value:
            result.append("%i %s read" % (value, key))

    return ", ".join(result) or "No reads"


def _key_by_tid_pos(record):
    return (record[0].tid, record[0].pos)


def check_fasta_file(filename):
    with open(filename) as handle:
        namecache = {}
        state, linelength, linelengthchanged = _NA, None, False
        for linenum, line in enumerate(handle, start=1):
            # Only \n is allowed as not all tools  handle \r
            line = line.rstrip("\n")

            if not line:
                if state in (_NA, _IN_WHITESPACE):
                    continue
                elif state == _IN_HEADER:
                    raise NodeError(
                        "Expected FASTA sequence, found empty line"
                        "\n    Filename = %r\n    Line = %r" % (filename, linenum)
                    )
                elif state == _IN_SEQUENCE:
                    state = _IN_WHITESPACE
                else:
                    assert False
            elif line.startswith(">"):
                if state in (_NA, _IN_SEQUENCE, _IN_WHITESPACE):
                    _validate_fasta_header(filename, linenum, line, namecache)
                    state = _IN_HEADER
                    linelength = None
                    linelengthchanged = False
                elif state == _IN_HEADER:
                    raise NodeError(
                        "Empty sequences not allowed\n"
                        "    Filename = %r\n    Line = %r" % (filename, linenum - 1)
                    )
                else:
                    assert False
            else:
                if state == _NA:
                    raise NodeError(
                        "Expected FASTA header, found %r\n"
                        "    Filename = %r\n    Line = %r" % (line, filename, linenum)
                    )
                elif state == _IN_HEADER:
                    _validate_fasta_line(filename, linenum, line)
                    linelength = len(line)
                    state = _IN_SEQUENCE
                elif state == _IN_SEQUENCE:
                    _validate_fasta_line(filename, linenum, line)
                    # If the length has changed, then that line must be the
                    # last line in the record, which may be shorter due to the
                    # sequence length. This is because the FAI index format
                    # expects that each line has the same length.
                    if linelengthchanged or (linelength < len(line)):
                        raise NodeError(
                            "Lines in FASTQ files must be of same "
                            "length\n    Filename = %r\n"
                            "    Line = %r" % (filename, linenum)
                        )
                    elif linelength != len(line):
                        linelengthchanged = True
                elif state == _IN_WHITESPACE:
                    raise NodeError(
                        "Empty lines not allowed in sequences\n"
                        "    Filename = %r\n    Line = %r" % (filename, linenum)
                    )
                else:
                    assert False

        if state == _NA:
            raise NodeError(
                "File does not contain any sequences:\n    Filename = %r" % (filename,)
            )
        elif state == _IN_HEADER:
            raise NodeError(
                "File ends with an empty sequence:\n    Filename = %r" % (filename,)
            )


# Standard nucleotides + UIPAC codes
_VALID_CHARS_STR = "ACGTNRYSWKMBDHV"
_VALID_CHARS = frozenset(_VALID_CHARS_STR.upper() + _VALID_CHARS_STR.lower())
_NA, _IN_HEADER, _IN_SEQUENCE, _IN_WHITESPACE = range(4)


def _validate_fasta_header(filename, linenum, line, cache):
    name = line.split(" ", 1)[0][1:]
    if not name:
        raise NodeError(
            "FASTA sequence must have non-empty name\n"
            "    Filename = %r\n    Line = %r\n" % (filename, linenum)
        )
    elif not _RE_REF_NAME.match(name):
        raise NodeError(
            "Invalid name for FASTA sequence: %r\n"
            "    Filename = %r\n    Line = %r\n" % (name, filename, linenum)
        )
    elif name in cache:
        raise NodeError(
            "FASTA sequences have identical name\n"
            "    Filename = %r\n    Name = %r\n"
            "    Line 1 = %r\n    Line 2 = %r\n"
            % (filename, name, linenum, cache[name])
        )
    cache[name] = linenum


_RE_REF_NAME = re.compile("[!-()+-<>-~][!-~]*")


def _validate_fasta_line(filename, linenum, line):
    invalid_chars = frozenset(line) - _VALID_CHARS
    if invalid_chars:
        if invalid_chars == frozenset("\r"):
            raise NodeError(
                "FASTA file contains carriage-returns ('\\r')!\n"
                "Please convert file to unix format, using e.g. "
                "dos2unix.\n    Filename = %r\n" % (filename,)
            )

        raise NodeError(
            "FASTA sequence contains invalid characters\n"
            "    Filename = %r\n    Line = %r\n"
            "    Invalid characters = %r" % (filename, linenum, "".join(invalid_chars))
        )



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import copy
import os

import paleomix.pipelines.bam.paths as paths
from paleomix.common.fileutils import swap_ext
from paleomix.common.makefile import MakefileError
from paleomix.nodes.bowtie2 import Bowtie2Node
from paleomix.nodes.bwa import BWAAlgorithmNode, BWABacktrack, BWASampe, BWASamse
from paleomix.pipelines.bam.nodes import index_and_validate_bam
from paleomix.pipelines.bam.parts.reads import Reads

#
_TRIMMED_READS_CACHE = {}


class Lane:
    def __init__(self, config, prefix, record, name):
        self.name = name
        self.bams = {}
        self.reads = None
        self.options = copy.deepcopy(record["Options"])
        self.tags = tags = copy.deepcopy(record["Tags"])
        self.tags["PG"] = self.tags["PG"].lower()
        self.folder = os.path.join(
            config.destination,
            tags["Target"],
            prefix["Name"],
            tags["SM"],
            tags["LB"],
            tags["PU"],
            tags["Folder"],
        )

        self._init_reads(config, record)
        self._init_unaligned_lane(config, prefix)

    def _init_reads(self, config, record):
        key = tuple(self.tags[key] for key in ("Target", "SM", "LB", "PU", "DS"))
        if key not in _TRIMMED_READS_CACHE:
            _TRIMMED_READS_CACHE[key] = Reads(
                config, record, record["Options"]["QualityOffset"]
            )
        self.reads = _TRIMMED_READS_CACHE[key]

    def _init_unaligned_lane(self, config, prefix):
        prefix_key = "Nodes:%s" % (self.options["Aligners"]["Program"],)

        for key, input_filename in self.reads.files.items():
            # Common parameters between BWA / Bowtie2
            output_filename = os.path.join(self.folder, "%s.bam" % (key.lower(),))

            alignment_node = self._build_alignment_node(
                config=config,
                prefix=prefix,
                parameters={
                    "input_file": input_filename,
                    "output_file": output_filename,
                    "prefix": prefix["Path"],
                    "reference": prefix["Reference"],
                    "dependencies": self.reads.nodes + prefix[prefix_key],
                },
            )

            self.bams[key] = {
                output_filename: self._finalize_nodes(
                    config=config,
                    prefix=prefix,
                    node=alignment_node,
                )
            }

    def _build_alignment_node(self, config, prefix, parameters):
        if self.options["Aligners"]["Program"] == "BWA":
            algorithm = self.options["Aligners"]["BWA"]["Algorithm"].lower()
            parameters["threads"] = config.bwa_max_threads
            parameters["algorithm"] = algorithm

            if algorithm == "backtrack":
                return self._build_bwa_backtrack(
                    config=config, prefix=prefix, parameters=parameters
                )
            elif algorithm in ("mem", "bwasw"):
                return self._build_bwa_algorithm(
                    config=config, prefix=prefix, parameters=parameters
                )
            else:
                raise NotImplementedError("BWA %r not implemented!" % (algorithm,))

        elif self.options["Aligners"]["Program"] == "Bowtie2":
            return self._build_bowtie2(
                config=config, prefix=prefix, parameters=parameters
            )
        else:
            raise NotImplementedError(
                "Aligner %r not implemented!" % (self.options["Aligners"]["Program"],)
            )

    ###########################################################################
    ###########################################################################
    # Construction of mapping nodes

    def _build_bwa_backtrack(self, config, prefix, parameters):
        if paths.is_paired_end(parameters["input_file"]):
            return self._build_bwa_backtrack_pe(
                config=config, prefix=prefix, parameters=parameters
            )
        else:
            return self._build_bwa_backtrack_se(
                config=config, prefix=prefix, parameters=parameters
            )

    def _build_bwa_backtrack_aln(self, parameters, input_file, output_file):
        options = dict(self.options["Aligners"]["BWA"])
        if not self.options["Aligners"]["BWA"]["UseSeed"]:
            options["-l"] = 2**16 - 1

        if self.reads.quality_offset in (64, "Solexa"):
            options["-I"] = None

        return BWABacktrack(
            input_file=input_file,
            output_file=output_file,
            threads=parameters["threads"],
            prefix=parameters["prefix"],
            reference=parameters["reference"],
            mapping_options=options,
            dependencies=parameters["dependencies"],
        )

    def _build_bwa_backtrack_se(self, config, prefix, parameters):
        input_file_fq = parameters.pop("input_file")
        output_file_bam = parameters.pop("output_file")
        output_file_sai = swap_ext(output_file_bam, ".sai")

        aln_node = self._build_bwa_backtrack_aln(
            parameters=parameters, input_file=input_file_fq, output_file=output_file_sai
        )

        return BWASamse(
            input_file_fq=input_file_fq,
            input_file_sai=output_file_sai,
            output_file=output_file_bam,
            prefix=parameters["prefix"],
            reference=parameters["reference"],
            # FIXME: There needs to be a way to set BWASamse options
            # mapping_options=self.options["Aligners"]["BWA"],
            cleanup_options=self._cleanup_options("BWA"),
            dependencies=aln_node,
        )

    def _build_bwa_backtrack_pe(self, config, prefix, parameters):
        template = parameters.pop("input_file")
        output_bam = parameters.pop("output_file")

        output_sai_1 = swap_ext(output_bam, "%i.sai" % (1,))
        aln_node_1 = self._build_bwa_backtrack_aln(
            parameters=parameters,
            input_file=template.format(Pair=1),
            output_file=output_sai_1,
        )

        output_sai_2 = swap_ext(output_bam, "%i.sai" % (2,))
        aln_node_2 = self._build_bwa_backtrack_aln(
            parameters=parameters,
            input_file=template.format(Pair=2),
            output_file=output_sai_2,
        )

        return BWASampe(
            input_file_sai_1=output_sai_1,
            input_file_sai_2=output_sai_2,
            input_file_fq_1=template.format(Pair=1),
            input_file_fq_2=template.format(Pair=2),
            output_file=output_bam,
            prefix=parameters["prefix"],
            reference=parameters["reference"],
            # FIXME: There needs to be a way to set BWASamse options
            # mapping_options=self.options["Aligners"]["BWA"],
            cleanup_options=self._cleanup_options("BWA"),
            dependencies=(aln_node_1, aln_node_2),
        )

    def _build_bwa_algorithm(self, config, prefix, parameters):
        if self.reads.quality_offset != 33:
            raise MakefileError(
                "Mapping with BWA using the %r algorithm currently does not support "
                "QualityOffsets other than 33; please convert your FASTQ if you wish "
                "to proceed." % (parameters["algorithm"],)
            )

        parameters = self._set_pe_input_files(parameters)
        parameters["mapping_options"] = self.options["Aligners"]["BWA"]
        parameters["cleanup_options"] = self._cleanup_options("BWA")

        return BWAAlgorithmNode(**parameters)

    def _build_bowtie2(self, config, prefix, parameters):
        parameters = self._set_pe_input_files(parameters)
        parameters["threads"] = config.bowtie2_max_threads
        parameters["index_format"] = prefix["IdxFmt:Bowtie2"]

        parameters["mapping_options"] = dict(self.options["Aligners"]["Bowtie2"])
        parameters["cleanup_options"] = self._cleanup_options("Bowtie2")

        if self.reads.quality_offset != 33:
            parameters["mapping_options"]["--phred64"] = None

        return Bowtie2Node(**parameters)

    def _cleanup_options(self, aligner):
        options = {
            # Add mate-score tag required by `samtools markdup`
            "--add-mate-score": None,
            "--rg-id": self.tags["ID"],
            "--rg": [
                "%s:%s" % (tag_name, self.tags[tag_name])
                for tag_name in ("SM", "LB", "PU", "PL", "PG", "DS")
            ],
            "-q": self.options["Aligners"][aligner]["MinQuality"],
        }

        if self.options["Aligners"][aligner]["FilterUnmappedReads"]:
            options["-F"] = "0x4"

        return options

    def _finalize_nodes(self, config, prefix, node):
        # Indexes are required for all files when calculating region statistics
        return index_and_validate_bam(
            config=config,
            prefix=prefix,
            node=node,
            create_index=bool(prefix.get("RegionsOfInterest")),
        )

    @staticmethod
    def _set_pe_input_files(parameters):
        parameters = dict(parameters)
        template = parameters.pop("input_file")
        if paths.is_paired_end(template):
            parameters["input_file_1"] = template.format(Pair=1)
            parameters["input_file_2"] = template.format(Pair=2)
        else:
            parameters["input_file_1"] = template
            parameters["input_file_2"] = None

        return parameters



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import collections
import os

from paleomix.common.fileutils import swap_ext
from paleomix.nodes.commands import CoverageNode, DepthHistogramNode, MergeCoverageNode
from paleomix.pipelines.bam.parts.summary import SummaryTableNode


def add_statistics_nodes(config, makefile, target):
    features = makefile["Options"]["Features"]

    nodes = []
    if features["Depths"]:
        nodes.extend(_build_depth(config, target, makefile["Prefixes"]))

    if features["Summary"] or features["Coverage"]:
        make_summary = features["Summary"]
        coverage = _build_coverage(config, target, make_summary)
        if make_summary:
            summary_node = _build_summary_node(config, makefile, target, coverage)
            nodes.append(summary_node)
        elif features["Coverage"]:
            nodes.extend(coverage["Nodes"])

    target.nodes.extend(nodes)


def _build_summary_node(config, makefile, target, coverage):
    coverage_by_label = _build_coverage_nodes(target)

    return SummaryTableNode(
        config=config,
        makefile=makefile,
        target=target,
        cov_for_lanes=coverage_by_label["Lanes"],
        cov_for_libs=coverage_by_label["Libraries"],
        dependencies=coverage["Nodes"],
    )


def _build_depth(config, target, prefixes):
    nodes = []
    for prefix in target.prefixes:
        for roi_name, roi_filename in _get_roi(prefix, name_prefix="."):
            ((input_file, dependencies),) = prefix.bams.items()

            output_filename = "%s.%s%s.depths" % (target.name, prefix.name, roi_name)
            output_fpath = os.path.join(config.destination, output_filename)

            nodes.append(
                DepthHistogramNode(
                    target_name=target.name,
                    input_file=input_file,
                    prefix=prefixes[prefix.name],
                    regions_file=roi_filename,
                    output_file=output_fpath,
                    dependencies=dependencies,
                )
            )

    return nodes


def _aggregate_for_prefix(cov, prefix, roi_name=None, into=None):
    prefix = _get_prefix_label(prefix, roi_name)
    results = {} if into is None else into
    for key, files_and_nodes in cov.items():
        if prefix is None or (key[0] == prefix):
            results.update(files_and_nodes)
    return results


def _build_coverage(config, target, make_summary):
    merged_nodes = []
    coverage = _build_coverage_nodes(target)
    for prefix in target.prefixes:
        for roi_name, _ in _get_roi(prefix):
            label = _get_prefix_label(prefix.name, roi_name)
            if not roi_name:
                postfix = prefix.name
            else:
                postfix = "%s.%s" % (prefix.name, roi_name)

            files_and_nodes = _aggregate_for_prefix(coverage["Libraries"], label)
            output_filename = os.path.join(
                config.destination, "%s.%s.coverage" % (target.name, postfix)
            )
            merged = MergeCoverageNode(
                input_files=list(files_and_nodes.keys()),
                output_file=output_filename,
                dependencies=list(files_and_nodes.values()),
            )

            merged_nodes.append(merged)

    files_and_nodes = _aggregate_for_prefix(coverage["Libraries"], None)
    if make_summary:
        files_and_nodes = _aggregate_for_prefix(
            coverage["Lanes"], None, into=files_and_nodes
        )

    all_nodes = []
    all_nodes.extend(files_and_nodes.values())
    all_nodes.extend(merged_nodes)

    coverage["Nodes"] = tuple(all_nodes)

    return coverage


def _build_coverage_nodes(target):
    coverage = {
        "Lanes": collections.defaultdict(dict),
        "Libraries": collections.defaultdict(dict),
    }

    cache = {}
    for prefix in target.prefixes:
        for roi_name, roi_filename in _get_roi(prefix):
            prefix_label = _get_prefix_label(prefix.name, roi_name)

            for sample in prefix.samples:
                for library in sample.libraries:
                    key = (prefix_label, target.name, sample.name, library.name)

                    for lane in library.lanes:
                        for bams in lane.bams.values():
                            bams = _build_coverage_nodes_cached(
                                bams, target.name, roi_name, roi_filename, cache
                            )

                            coverage["Lanes"][key].update(bams)

                    bams = _build_coverage_nodes_cached(
                        library.bams, target.name, roi_name, roi_filename, cache
                    )
                    coverage["Libraries"][key].update(bams)
    return coverage


def _build_coverage_nodes_cached(
    files_and_nodes, target_name, roi_name, roi_filename, cache
):
    output_ext = ".coverage"
    if roi_name:
        output_ext = ".%s.coverage" % roi_name

    coverages = {}
    for input_filename, node in files_and_nodes.items():
        output_filename = swap_ext(input_filename, output_ext)

        cache_key = (roi_filename, input_filename)
        if cache_key not in cache:
            cache[cache_key] = CoverageNode(
                input_file=input_filename,
                output_file=output_filename,
                target_name=target_name,
                regions_file=roi_filename,
                dependencies=node,
            )

        coverages[output_filename] = cache[cache_key]
    return coverages


def _get_roi(prefix, name_prefix=""):
    roi = [("", None)]
    for name, path in prefix.roi.items():
        roi.append((name_prefix + name, path))
    return roi


def _get_prefix_label(label, roi_name):
    if not roi_name:
        return label
    return "%s:%s" % (label, roi_name)



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#


import copy
import os

import pysam

import paleomix.common.fileutils as fileutils
import paleomix.common.utilities as utilities
from paleomix.common.formats.fasta import FASTA
from paleomix.common.formats.msa import MSA
from paleomix.node import Node, NodeError


class CollectSequencesNode(Node):
    def __init__(self, fasta_files, sequences, destination, dependencies=()):
        """
        fasta_files -- { taxon_name_1 : filename_1, ... }
        sequences   -- { interval_name_1, ... }
        """

        self._infiles = copy.deepcopy(fasta_files)
        self._sequences = utilities.safe_coerce_to_frozenset(sequences)
        self._destination = copy.copy(destination)
        self._outfiles = [
            os.path.join(destination, name + ".fasta") for name in self._sequences
        ]

        input_files = list(self._infiles.values())
        for filename in self._infiles.values():
            input_files.append(filename + ".fai")

        desc = "<CollectSequences: %i sequences from %i files -> '%s'>" % (
            len(self._sequences),
            len(self._infiles),
            self._destination,
        )
        Node.__init__(
            self,
            description=desc,
            input_files=input_files,
            output_files=self._outfiles,
            dependencies=dependencies,
        )

    def _setup(self, _config, _temp):
        for filename in self._infiles.values():
            with open(filename + ".fai") as handle:
                sequences = set()
                for line in handle:
                    sequences.add(line.split("\t", 1)[0])

                missing_sequences = list(self._sequences - sequences)
                if missing_sequences:
                    if len(missing_sequences) >= 4:
                        missing_sequences = missing_sequences[:3]
                        missing_sequences.append("...")

                    message = (
                        "FASTA file does not contain expected "
                        "sequences:\n  File =  %r\n  "
                        "Sequences = %s\n"
                    ) % (filename, ", ".join(missing_sequences))
                    raise NodeError(message)

    def _run(self, _config, temp):
        fasta_files = []
        for name, filename in sorted(self._infiles.items()):
            fasta_files.append((name, pysam.FastaFile(filename)))

        for sequence_name in sorted(self._sequences):
            filename = os.path.join(temp, sequence_name + ".fasta")
            with open(filename, "w") as out_handle:
                for sample, fasta_file in fasta_files:
                    sequence = fasta_file.fetch(sequence_name)
                    fasta = FASTA(sample, sequence_name, sequence)
                    fasta.write(out_handle)

    def _teardown(self, _config, temp):
        for destination in sorted(self._outfiles):
            source = fileutils.reroot_path(temp, destination)
            fileutils.move_file(source, destination)


class FilterSingletonsNode(Node):
    def __init__(self, input_file, output_file, filter_by, dependencies):
        self._input_file = input_file
        self._output_file = output_file
        self._filter_by = dict(filter_by)
        for to_filter, groups in self._filter_by.items():
            # The taxa to be filtered is implied to be part of the group,
            # but is not needed when actually carrying out the filtering
            groups = utilities.safe_coerce_to_frozenset(
                groups
            ) - utilities.safe_coerce_to_frozenset(to_filter)

            if not groups:
                raise RuntimeError(
                    "Singleton filtering must involve at least one other taxa"
                )
            self._filter_by[to_filter] = groups

        Node.__init__(
            self,
            description="filtering singleton nucleotides in %s" % (input_file,),
            input_files=[input_file],
            output_files=[output_file],
            dependencies=dependencies,
        )

    def _run(self, _config, temp):
        alignment = MSA.from_file(self._input_file)
        for to_filter, groups in self._filter_by.items():
            alignment = alignment.filter_singletons(to_filter, groups)

        temp_filename = fileutils.reroot_path(temp, self._output_file)
        with open(temp_filename, "w") as handle:
            alignment.to_file(handle)
        fileutils.move_file(temp_filename, self._output_file)



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import sys

import pysam

from paleomix.common.fileutils import fspath, open_ro
from paleomix.common.formats._common import FormatError
from paleomix.common.utilities import Immutable, TotallyOrdered, fragment, split_before


class FASTAError(FormatError):
    pass


class FASTA(TotallyOrdered, Immutable):
    def __init__(self, name, meta, sequence):
        if not (name and isinstance(name, str)):
            raise FASTAError("FASTA name must be a non-empty string")
        elif not (isinstance(meta, str) or (meta is None)):
            raise FASTAError("FASTA meta must be a string, or None")
        elif not isinstance(sequence, str):
            raise FASTAError("FASTA sequence must be a string")

        Immutable.__init__(self, name=name, meta=meta or "", sequence=sequence)

    def write(self, fileobj=sys.stdout):
        """Writes a FASTA record to fileobj, wrapping sequences at 60 chars"""
        name = self.name
        if self.meta:
            name = "%s %s" % (name, self.meta)

        fileobj.write(">%s\n%s\n" % (name, "\n".join(fragment(60, self.sequence))))

    @classmethod
    def from_lines(cls, lines):
        """Parses FASTA sequences found in a sequence of lines, and returns
        a tuple for each FASTA record: ((name, meta-information), sequence)
        No assumptions are made about the line-lengths."""
        lines = (line.rstrip() for line in lines)
        for record in split_before(lines, lambda v: v.startswith(">")):
            name = record[0]
            if (not name.startswith(">")) or (len(name) == 1):
                raise FASTAError("Unnamed FASTA record")
            elif len(record) == 1:
                raise FASTAError(
                    "FASTA record does not contain sequence: %s" % (name[1:],)
                )

            # Split out any meta information
            name_and_meta = name[1:].split(None, 1)
            if len(name_and_meta) < 2:
                name_and_meta.append("")
            name, meta = name_and_meta

            yield FASTA(name=name, meta=meta, sequence="".join(record[1:]))

    @classmethod
    def from_file(cls, filename):
        """Reads an unindexed FASTA file, returning a sequence of
        tuples containing the name and sequence of each entry in
        the file. The FASTA file may be GZIP/BZ2 compressed."""
        with open_ro(filename) as fasta_file:
            yield from FASTA.from_lines(fasta_file)

    @classmethod
    def index_and_collect_contigs(cls, filename):
        """Creates an index (.fai; if it does not already exist) for a FASTA file using
        pysam, and returns a dictionary of {contig: length} listed in that file.
        """
        # If an index does not already exist, then it is created automatically
        with pysam.FastaFile(fspath(filename)) as handle:
            return dict(zip(handle.references, handle.lengths))

    def __lt__(self, other):
        if not isinstance(other, FASTA):
            return NotImplemented

        return (self.name, self.meta, self.sequence) < (
            other.name,
            other.meta,
            other.sequence,
        )

    def __hash__(self):
        return hash((self.name, self.meta, self.sequence))

    def __repr__(self):
        return "FASTA(%r, %r, %r)" % (self.name, self.meta, self.sequence)



#!/usr/bin/python3
#
# Copyright (c) 2016 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import paleomix
import paleomix.common.logging
from paleomix.common.argparse import SUPPRESS, ArgumentParser
from paleomix.common.procs import max_processes

_RUN_USAGE = """%(prog)s [..] <database.tar> <samples.txt> [destination]
       %(prog)s [..] <database.tar> <sample.bam> [destination]
       %(prog)s [..] <database.tar> <nuclear.bam> <mitochondrial.bam> <destination>
"""

_DEFAULT_CONFIG_FILES = [
    "/etc/paleomix/zonkey.ini",
    "~/.paleomix/zonkey.ini",
]


def build_parser():
    parser = ArgumentParser(prog="paleomix zonkey")

    subparsers = parser.add_subparsers(dest="command", metavar="command")
    add_copy_example_command(subparsers)
    add_mito_command(subparsers)

    return parser, add_run_command(subparsers)


def add_copy_example_command(subparsers):
    parser = subparsers.add_parser("example", help="Create example project")

    parser.add_argument(
        "database", help="Zonkey database file (uncompressed)", metavar="database.tar"
    )

    parser.add_argument(
        "destination",
        default=".",
        nargs="?",
        help="Destination folder for example data.",
    )


def add_mito_command(subparsers):
    parser = subparsers.add_parser(
        "mito", help="Create phylo pipeline project for mt alignments"
    )

    parser.add_argument(
        "database", help="Zonkey database file (uncompressed)", metavar="database.tar"
    )

    parser.add_argument(
        "destination",
        default=".",
        nargs="?",
        help="Destination folder for project template.",
    )


def add_run_command(subparsers):
    parser = subparsers.add_parser(
        "run",
        aliases=("dryrun",),
        help="Run pipeline on provided makefiles",
        usage=_RUN_USAGE,
        prog="paleomix zonkey run",
        default_config_files=_DEFAULT_CONFIG_FILES,
    )

    parser.add_argument(
        "database", help="Zonkey database file (uncompressed)", metavar="database.tar"
    )

    parser.add_argument("files", help="See usage", nargs="+")

    group = parser.add_argument_group("Program options")
    group.add_argument(
        "--downsample-to",
        type=int,
        default=1000000,
        help="Number of reads to use for analyses; if 0, no downsampling is performed",
    )
    group.add_argument(
        "--admixture-replicates",
        type=int,
        default=1,
        help="Number of admixture replicates to run, before "
        "the result with the highest likelihood",
    )
    group.add_argument(
        "--treemix-k",
        type=int,
        default=0,
        help="Value passed to treemix's -k option; number of "
        "SNPs per block for estimation of the covariance "
        "matrix. If set to 0, a value will be estimated "
        "assuming an even distribution of SNPs",
    )
    group.add_argument(
        "--treemix-outgroup",
        default="",
        type=lambda value: tuple(_f for _f in sorted(value.split(",")) if _f),
        help="Comma-seperated list of samples to use as the "
        "outgroup when running TreeMix; note that these "
        "must form a monophyletic clade, or TreeMix will "
        "not execute.",
    )

    group.add_argument(
        "--admixture-only",
        help=SUPPRESS,
        default=False,
        action="store_true",
    )

    paleomix.common.logging.add_argument_group(parser)

    group = parser.add_argument_group("Pipeline")
    group.add_argument(
        "--dry-run",
        action="store_true",
        default=False,
        help="If passed, only a dry-run in performed, the "
        "dependency tree is printed, and no tasks are "
        "executed.",
    )
    group.add_argument(
        "--max-threads",
        type=int,
        default=max(2, max_processes()),
        help="Maximum number of threads to use",
    )
    group.add_argument(
        "--list-input-files",
        action="store_true",
        default=False,
        help="List all input files used by pipeline for the "
        "makefile(s), excluding any generated by the "
        "pipeline itself.",
    )
    group.add_argument(
        "--list-output-files",
        action="store_true",
        default=False,
        help="List all output files generated by pipeline for the makefile(s).",
    )
    group.add_argument(
        "--list-executables",
        action="store_true",
        default=False,
        help="List all executables required by the pipeline, "
        "with version requirements (if any).",
    )

    # Removed options
    parser.add_argument("--progress-ui", help=SUPPRESS)
    parser.add_argument("--ui-colors", help=SUPPRESS)

    return parser



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
from unittest.mock import patch

from paleomix.common.formats.fasta import FASTA
from paleomix.common.formats.msa import MSA
from paleomix.common.formats.phylip import interleaved_phy

_MSA_SHORT_SEQUENCES = MSA(
    [FASTA("seq1", None, "ACGTTGATAACCAGG"), FASTA("seq2", None, "TGCAGAGTACGACGT")]
)

_MSA_MEDIUM_SEQUENCES = MSA(
    [
        FASTA("seq1", None, "ACGTTGATAACCAGGAGGGATTCGCGATTGGTGGTAACGTAGCC"),
        FASTA("seq2", None, "TGCAGAGTACGACGTCTCCTAGATCCTGGACAATTTAAACCGAA"),
    ]
)

_MSA_LONG_SEQUENCES = MSA(
    [
        FASTA(
            "seq1",
            None,
            "CGGATCTGCTCCTCCACTGGCCACGTTTACTGTCCCCCAACCGTT"
            "CGTCCCGACCTAGTTATACTTCTTAGCAAGGTGTAAAACCAGAGATTGAGGTTATAACG"
            "TTCCTAATCAGTTATTAAATTACCGCGCCCCGACAG",
        ),
        FASTA(
            "seq2",
            None,
            "AGTTGAAGAGGCGGAACGTTTGTAAACCGCGCTAACGTAGTTCTA"
            "CAACCAGCCACCCGGTTCGAAGGAACAACTGGTCGCCATAATTAGGCGAAACGATAGTG"
            "CACTAAGGTCAGGTGCGCCCCTGTAAATAATTAGAT",
        ),
    ]
)

_MSA_MEDIUM_NAMES = MSA(
    [
        FASTA("A_really_long_sequence", None, "ACGTTGATAACCAGG"),
        FASTA("Another_real_long_one!", None, "TGCAGAGTACGACGT"),
    ]
)

_MSA_LONG_NAMES = MSA(
    [
        FASTA(
            "A_really_long_sequence_name_that_is_in_fact_too_long",
            None,
            "ACGTTGATAACCAGG",
        ),
        FASTA(
            "Another_really_long_sequence_name_that_is_too_long",
            None,
            "TGCAGAGTACGACGT",
        ),
    ]
)


###############################################################################
###############################################################################
# Tests of 'interleaved_phy'


def test_interleaved_phy__short_sequences():
    expected = """2 44

seq1        ACGTTGATAA  CCAGGAGGGA  TTCGCGATTG  GTGGTAACGT  AGCC
seq2        TGCAGAGTAC  GACGTCTCCT  AGATCCTGGA  CAATTTAAAC  CGAA"""
    assert interleaved_phy(_MSA_MEDIUM_SEQUENCES) == expected


def test_interleaved_phy__multi_line_sequences():
    expected = """2 140

seq1        CGGATCTGCT  CCTCCACTGG  CCACGTTTAC  TGTCCCCCAA  CCGTTCGTCC
seq2        AGTTGAAGAG  GCGGAACGTT  TGTAAACCGC  GCTAACGTAG  TTCTACAACC

CGACCTAGTT  ATACTTCTTA  GCAAGGTGTA  AAACCAGAGA  TTGAGGTTAT  AACGTTCCTA
AGCCACCCGG  TTCGAAGGAA  CAACTGGTCG  CCATAATTAG  GCGAAACGAT  AGTGCACTAA

ATCAGTTATT  AAATTACCGC  GCCCCGACAG
GGTCAGGTGC  GCCCCTGTAA  ATAATTAGAT"""
    assert interleaved_phy(_MSA_LONG_SEQUENCES) == expected


def test_interleaved_phy__with_flag():
    expected = """2 15 I

seq1        ACGTTGATAA  CCAGG
seq2        TGCAGAGTAC  GACGT"""
    assert interleaved_phy(_MSA_SHORT_SEQUENCES, add_flag=True) == expected


def test_interleaved_phy__medium_names():
    expected = """2 15

A_really_long_sequence  ACGTTGATAA  CCAGG
Another_real_long_one!  TGCAGAGTAC  GACGT"""
    assert interleaved_phy(_MSA_MEDIUM_NAMES) == expected


def test_interleaved_phy__long_names():
    expected = """2 15

A_really_long_sequence_name_th      ACGTTGATAA  CCAGG
Another_really_long_sequence_n      TGCAGAGTAC  GACGT"""
    assert interleaved_phy(_MSA_LONG_NAMES) == expected


def test_sequentual_phy__different_length_names_1():
    msa = MSA(
        [
            FASTA("A_short_name", None, "ACGTTGATAACCAGG"),
            FASTA(
                "Another_really_long_sequence_name_that_is_too_long",
                None,
                "TGCAGAGTACGACGT",
            ),
        ]
    )
    expected = """2 15

A_short_name                        ACGTTGATAA  CCAGG
Another_really_long_sequence_n      TGCAGAGTAC  GACGT"""
    assert interleaved_phy(msa) == expected


def test_sequentual_phy__different_length_names_2():
    msa = MSA(
        [
            FASTA("Burchelli_4", None, "ACGTTGATAACCAGG"),
            FASTA("Donkey", None, "TGCAGAGTACGACGT"),
        ]
    )
    expected = """2 15

Burchelli_4             ACGTTGATAA  CCAGG
Donkey                  TGCAGAGTAC  GACGT"""
    assert interleaved_phy(msa) == expected


def test_interleaved_phy__different_lengths():
    with patch("paleomix.common.formats.msa.MSA.validate", wrap=MSA.validate) as mock:
        interleaved_phy(_MSA_MEDIUM_NAMES)

    assert len(mock.mock_calls) == 1



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Various functions relating to DNA sequence manipulation."""

import itertools

# Pairs of complementary bases and ambiguous basees
_COMPL = ["AT", "CG", "NN", "RY", "KM", "SS", "WW", "BV", "DH", "XX"]
_COMPL_TABLE = ["N"] * 256
for _a, _b in _COMPL:
    # Complement both upper/lower-case bases
    for _func in (str.upper, str.lower):
        _COMPL_TABLE[ord(_func(_a))] = _func(_b)
        _COMPL_TABLE[ord(_func(_b))] = _func(_a)
_COMPL_TABLE = "".join(_COMPL_TABLE)


# Table of nt codes (IUPAC codes) used to encode (ambiguous) bases:
#   Nomenclature for incompletely specified bases in nucleic acid sequences.
#   Recommendations 1984. J Biol Chem. 1986 Jan 5;261(1):13-7.
#   PubMed PMID: 2416744.
NT_CODES = [
    ["A", "A"],
    ["C", "C"],
    ["G", "G"],
    ["T", "T"],
    ["N", "N"],
    ["R", "AG"],
    ["Y", "CT"],
    ["K", "GT"],
    ["M", "AC"],
    ["S", "CG"],
    ["W", "AT"],
    ["B", "CGT"],
    ["D", "AGT"],
    ["H", "ACT"],
    ["V", "ACG"],
    ["N", "ACGT"],
]

_NT_CODES_TABLE = {}
for _abr, _nts in NT_CODES:
    _NT_CODES_TABLE[frozenset(_nts)] = _abr
    _NT_CODES_TABLE[frozenset(_nts + ",")] = _abr

NT_CODES = dict(NT_CODES)


def complement(sequence):
    """Returns the complement of a DNA sequence (string)."""
    return sequence.translate(_COMPL_TABLE)


def reverse_complement(sequence):
    """Returns the reverse complement of a DNA sequence."""
    return complement(sequence)[::-1]


def encode_genotype(nucleotides):
    """Parses a string representing a set of nucleotides observed at a loci,
    and returns the corresponding IUPAC code. Commas are allowed, but are
    simply ignored if found in the string. Does not handle lower-case
    nucleotides, due to lack of clear criteria for mixed case input.
    See e.g. http://www.ebi.ac.uk/2can/tutorials/aa.html"""
    try:
        return _NT_CODES_TABLE[frozenset(nucleotides)]
    except KeyError:
        raise ValueError(nucleotides)


def split(sequence, split_by="123"):
    """Splits a sequence by position, as specified by the 'split_by' parameter. By
    default, the function will split by codon position, and return a dictionary
    containing the keys '1', '2' and '3'.

    The 'split_by' parameter may contain any non-zero number of values, which must
    however be hashable. If a value is specified multiple times, then those positions
    are interleaved (e.g. split_by = "112" returns the first two positions in a codon
    as one sequence, as well as the last positions as one sequence."""
    if not split_by:
        raise ValueError("No split_by specified")

    results = dict((key, []) for key in split_by)
    keys = itertools.chain(itertools.cycle(split_by))
    for key, nucleotide in zip(keys, sequence):
        results[key].append(nucleotide)

    for key in results:
        results[key] = "".join(results[key])

    return results



#!/usr/bin/python3
#
# Copyright (c) 2013 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import errno
import itertools
import logging
import os
import time

import coloredlogs

_LOG_LEVELS = {
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "debug": logging.DEBUG,
}


_CONSOLE_MESSAGE_FORMAT = "%(asctime)s %(levelname)s %(message)s"
_CONSOLE_DATE_FORMAT = "%H:%M:%S"
_FILE_MESSAGE_FORMAT = "%(asctime)s %(name)s %(levelname)s %(message)s"

_LOG_ENABLED = False


def initialize_console_logging():
    global _LOG_ENABLED

    if not _LOG_ENABLED:
        coloredlogs.install(fmt=_CONSOLE_MESSAGE_FORMAT, datefmt=_CONSOLE_DATE_FORMAT)
        _LOG_ENABLED = True


def initialize(log_level="error", log_file=None, name="paleomix"):
    initialize_console_logging()

    log_level = _LOG_LEVELS[log_level.lower()]
    root = logging.getLogger()
    root.setLevel(logging.NOTSET)

    if log_file:
        handler = logging.FileHandler(log_file)
        handler.setFormatter(logging.Formatter(_FILE_MESSAGE_FORMAT))
        handler.setLevel(log_level)
        root.addHandler(handler)
    elif name:
        template = "%s.%s_%%02i.log" % (name, time.strftime("%Y%m%d_%H%M%S"))
        handler = LazyLogfile(template, log_level=log_level)
        handler.setFormatter(logging.Formatter(_FILE_MESSAGE_FORMAT))
        handler.setLevel(log_level)
        root.addHandler(handler)


def add_argument_group(parser, default="error"):
    """Adds an option-group to an OptionParser object, with options
    pertaining to logging. Note that 'initialize' expects the config
    object to have these options."""
    group = parser.add_argument_group("Logging")
    group.add_argument(
        "--log-file",
        default=None,
        help="Write log-messages to this file.",
    )
    group.add_argument(
        "--log-level",
        default=default,
        choices=("info", "warning", "error", "debug"),
        type=str.lower,
        help="Log messages at the specified level. If a filename is not specified with "
        "--log-file, PALEOMIX will automatically create a log file in the current "
        "directory should events be logged at this level or above",
    )


def get_logfiles():
    root = logging.getLogger()
    for handler in root.handlers:
        if isinstance(handler, logging.FileHandler) and handler.stream:
            yield handler.baseFilename


class LazyLogfile(logging.FileHandler):
    def __init__(self, template, log_level):
        logging.FileHandler.__init__(self, template, delay=True)
        # Use absolute path for template
        self._template = self.baseFilename
        self._log_level = log_level

    def emit(self, record):
        # Don't try to log self-emitted log events, to avoid recursive loops
        if record.name != __name__:
            super().emit(record)

    def _open(self):
        """Try to open a new logfile, taking steps to ensure that
        existing logfiles using the same template are not clobbered."""
        flags = os.O_WRONLY | os.O_CREAT | os.O_EXCL

        for start in itertools.count(start=1):
            filename = self._template % (start,)
            os.makedirs(os.path.dirname(filename), exist_ok=True)

            try:
                stream = os.fdopen(os.open(filename, flags), "w")
                logger = logging.getLogger(__name__)
                level_name = logging.getLevelName(self._log_level).lower()
                logger.info("Saving %s logs to %r", level_name, filename)

                self.baseFilename = filename
                return stream
            except OSError as error:
                if error.errno != errno.EEXIST:
                    raise



#!/usr/bin/python3
#
# Copyright (c) 2013 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import glob
import os
import re

import paleomix.common.fileutils as fileutils
import paleomix.common.versions as versions
from paleomix.atomiccmd.builder import (
    AtomicCmdBuilder,
    AtomicMPICmdBuilder,
)
from paleomix.node import CommandNode
from paleomix.nodegraph import FileStatusCache

EXAML_VERSION = versions.Requirement(
    call=("examl", "-version"),
    search=r"version (\d+)\.(\d+)\.(\d+)",
    checks=versions.GE(3, 0, 21),
)

PARSER_VERSION = versions.Requirement(
    call=("parse-examl", "-h"),
    search=r"version (\d+)\.(\d+)\.(\d+)",
    checks=versions.GE(3, 0, 21),
)


class ExaMLParserNode(CommandNode):
    def __init__(self, input_alignment, input_partition, output_file, dependencies=()):
        """
        Arguments:
        input_alignment  -- An alignment file in a format readable by RAxML.
        input_partition  -- A set of partitions in a format readable by RAxML.
        output_filename  -- Filename for the output binary sequence."""

        command = AtomicCmdBuilder("parse-examl", set_cwd=True)

        command.set_option("-s", "%(TEMP_OUT_ALN)s")
        command.set_option("-q", "%(TEMP_OUT_PART)s")
        # Output file will be named output.binary, and placed in the CWD
        command.set_option("-n", "output")

        # Substitution model
        command.set_option("-m", "DNA", fixed=False)

        command.set_kwargs(  # Auto-delete: Symlinks
            TEMP_OUT_PART=os.path.basename(input_partition),
            TEMP_OUT_ALN=os.path.basename(input_alignment),
            # Input files, are not used directly (see below)
            IN_ALIGNMENT=input_alignment,
            IN_PARTITION=input_partition,
            # Final output file, are not created directly
            OUT_BINARY=output_file,
            CHECK_EXAML=PARSER_VERSION,
        )

        CommandNode.__init__(
            self,
            command=command.finalize(),
            description="pre-parsing %s for ExaML" % input_alignment,
            dependencies=dependencies,
        )

        self._symlinks = [
            os.path.abspath(input_alignment),
            os.path.abspath(input_partition),
        ]
        self._output_file = os.path.basename(output_file)

    def _setup(self, config, temp):
        CommandNode._setup(self, config, temp)

        # Required to avoid the creation of files outside the temp folder
        for filename in self._symlinks:
            source = os.path.abspath(filename)
            destination = os.path.join(temp, os.path.basename(filename))

            os.symlink(source, destination)

    def _teardown(self, config, temp):
        os.remove(os.path.join(temp, "RAxML_info.output"))

        source = os.path.join(temp, "output.binary")
        destination = fileutils.reroot_path(temp, self._output_file)
        fileutils.move_file(source, destination)

        CommandNode._teardown(self, config, temp)


class ExaMLNode(CommandNode):
    def __init__(
        self,
        input_binary,
        initial_tree,
        output_template,
        model="GAMMA",
        threads=1,
        dependencies=(),
    ):
        """
        Arguments:
        input_binary  -- A binary alignment file in a format readable by ExaML.
        output_template  -- A template string used to construct final filenames. Should
                            consist of a full path, including a single '%s', which is
                            replaced with the variable part of RAxML output files (e.g.
                            'info', 'bestTree', ...).

                            Example destination: '/disk/project/SN013420.RAxML.%s'
                            Example output:      '/disk/project/SN013420.RAxML.bestTree'
        """

        # TODO: Make MPIParams!
        command = AtomicMPICmdBuilder("examl", threads=threads)

        # Ensures that output is saved to the temporary directory
        command.set_option("-w", "%(TEMP_DIR)s")

        command.set_option("-s", "%(IN_ALN)s")
        command.set_option("-t", "%(IN_TREE)s")
        command.set_option("-n", "Pypeline")

        command.set_kwargs(
            IN_ALN=input_binary,
            IN_TREE=initial_tree,
            # Final output files, are not created directly
            OUT_INFO=output_template % "info",
            OUT_BESTTREE=output_template % "result",
            OUT_BOOTSTRAP=output_template % "log",
            # Only generated by newer versions of ExaML
            TEMP_OUT_MODELFILE=os.path.basename(output_template % "modelFile"),
            CHECK_EXAML=EXAML_VERSION,
        )

        # Use the GAMMA model of NT substitution by default
        command.set_option("-m", model)

        self._dirname = os.path.dirname(output_template)
        self._template = os.path.basename(output_template)

        threads_txt = ""
        if threads > 1:
            threads_txt = " using %i threads"

        CommandNode.__init__(
            self,
            command=command.finalize(),
            description="running ExaML on %s%s" % (input_binary, threads_txt),
            threads=threads,
            dependencies=dependencies,
        )

    def _create_temp_dir(self, _config):
        """Called by 'run' in order to create a temporary folder.
        To allow restarting from checkpoints, we use a fixed folder
        determined by the output_template."""
        temp = os.path.join(self._dirname, self._template % ("temp",))
        fileutils.make_dirs(temp)
        return temp

    def _setup(self, config, temp):
        CommandNode._setup(self, config, temp)

        # The temp folder may contain old files:
        # Remove old pipes to prevent failure at _teardown
        for pipe_fname in glob.glob(os.path.join(temp, "pipe*")):
            fileutils.try_remove(pipe_fname)
        # ExaML refuses to overwrite old info files
        fileutils.try_remove(os.path.join(temp, "ExaML_info.Pypeline"))

        # Resume from last checkpoint, if one such was generated
        checkpoints = glob.glob(os.path.join(temp, "ExaML_binaryCheckpoint.Pypeline_*"))
        if not checkpoints:
            return

        cache = FileStatusCache()
        if not cache.are_files_outdated(self.input_files, checkpoints):
            checkpoints.sort(key=lambda fname: int(fname.rsplit("_", 1)[-1]))

            # FIXME: Less hacky solution to modifying AtomicCmds needed
            self._command._command.append("-R")
            self._command._command.append(checkpoints[-1])
        else:
            for fpath in checkpoints:
                fileutils.try_remove(fpath)

    def _teardown(self, config, temp):
        for filename in os.listdir(temp):
            match = re.match("ExaML_(.*).Pypeline", filename)
            if match:
                if "binaryCheckpoint" in match.groups():
                    os.remove(os.path.join(temp, filename))
                else:
                    source = os.path.join(temp, filename)
                    destination = os.path.join(temp, self._template % match.groups())

                    fileutils.move_file(source, destination)

        CommandNode._teardown(self, config, temp)



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import random
import re

from paleomix.common.fileutils import move_file, reroot_path
from paleomix.node import Node, NodeError


class PHYLIPBootstrapNode(Node):
    """Generates a bootstrap alignment for a partition PHYLIP file;

    Note that only the PHYLIP / partitions format produced by the Node
    FastaToPartitionedInterleavedPhyNode is supported, in addition to the
    formats produced by RAxMLReduceNode.

    Parameters:
      -- input_alignment  - The input alignment file in PHYLIP format
      -- input_partition  - The input partition file in RAxML format
      -- output_alignment - The output alignment file in PHYLIP format
                            The simple (RAxML like) sequential format is used.
      -- seed             - RNG seed for selecting alignment columns."""

    def __init__(
        self,
        input_alignment,
        input_partition,
        output_alignment,
        seed=None,
        dependencies=(),
    ):
        self._input_phy = input_alignment
        self._input_part = input_partition
        self._output_phy = output_alignment
        self._seed = seed

        Node.__init__(
            self,
            description="creating bootstraps from %s" % (input_alignment,),
            input_files=(input_alignment, input_partition),
            output_files=(output_alignment,),
            dependencies=dependencies,
        )

    def _run(self, _config, temp):
        if self._seed is not None:
            rng = random.Random(self._seed)
        partitions = _read_partitions(self._input_part)
        header, names, sequences = _read_sequences(self._input_phy)
        bootstraps = self._bootstrap_sequences(sequences, partitions, rng)

        temp_fpath = reroot_path(temp, self._output_phy)
        with open(temp_fpath, "w") as output_phy:
            output_phy.write(header)

            for name, fragments in zip(names, bootstraps):
                output_phy.write(name)
                output_phy.write(" ")
                for sequence in fragments:
                    output_phy.write(sequence)
                output_phy.write("\n")

        move_file(temp_fpath, self._output_phy)

    @classmethod
    def _bootstrap_sequences(cls, sequences, partitions, rng):
        final_partitions = [[] for _ in sequences]
        for start, end in partitions:
            # Convert alignment to columns, and randomly select among those
            columns = list(zip(*(sequence[start:end] for sequence in sequences)))
            bootstrap_partition = (rng.choice(columns) for _ in columns)

            # Convert randomly selected columns back into sequences
            for dest, partition in zip(final_partitions, zip(*bootstrap_partition)):
                dest.append("".join(partition))

        return final_partitions


_RE_PARTITION = re.compile(r"^[A-Z]+, [^ ]+ = (\d+)-(\d+)$")
_RE_PARTITION_SINGLE = re.compile(r"^[A-Z]+, [^ ]+ = (\d+)$")


def _read_partitions(filename):
    """Read a partition file, as produced by the pipeline itself, and
    returns a list of tuples containing the (start, end) coordinates;
    each line is expected to follow the following format:

    DNA, Name = Start-End

    Multiple regions, or skips are not supported."""
    partitions = []
    with open(filename) as handle:
        for line_num, line in enumerate(handle):
            result = _RE_PARTITION.match(line.rstrip())
            if result:
                start, end = result.groups()
            else:
                result = _RE_PARTITION_SINGLE.match(line.rstrip())
                if not result:
                    message = (
                        "Line %i in partitions file does not follow "
                        "expected format:\n"
                        "  Expected, either = 'DNA, Name = Start-End'\n"
                        "                or = 'DNA, Name = Start'\n"
                        "  Found = %r"
                    ) % (line_num, line.rstrip())
                    raise NodeError(message)
                (start,) = result.groups()
                end = start

            partitions.append((int(start) - 1, int(end)))
    return partitions


def _read_sequences(filename):
    """Collects the sequences from a PHYLIP file, and returns the header,
    the names of the sequences, and the sequences themselves. The parser
    supports interleaved sequences (as produced by the pipeline), or simple
    sequential (each paired name and sequence on a single line) as produced
    by RAxML's reduce functionality. PHYLIP files containing multiple entries
    are not supported."""
    line, header = " ", None
    with open(filename) as handle:
        # Find header
        num_sequences = num_bases = 0
        while line:
            line = handle.readline()
            if line.strip():
                header = line
                num_sequences, num_bases = map(int, line.split())
                break

        names = [None for _ in range(num_sequences)]
        sequences = [[] for _ in range(num_sequences)]

        line_num = 0
        while line:
            line = handle.readline()
            line_strip = line.strip()
            if line_strip:
                # The first N sequences are expected to contain sample names
                index = line_num % num_sequences
                if line_num < num_sequences:
                    name, line_strip = line_strip.split(None, 1)
                    names[index] = name

                sequences[index].extend(line_strip.split())
                line_num += 1

    if len(sequences) != num_sequences:
        message = (
            "Expected %i sequences, but found %i in PHYLIP file:\n    Filename = %r"
        ) % (num_sequences, len(sequences), filename)
        raise NodeError(message)

    for index, fragments in enumerate(sequences):
        sequences[index] = "".join(fragments)
        if len(sequences[index]) != num_bases:
            raise NodeError(
                "Expected %ibp sequences, found %ibp sequence for %r\n Filename = %r"
                % (num_bases, len(sequences[index]), names[index], filename)
            )

    return header, names, sequences



#!/usr/bin/python3
#
# Copyright (c) 2016 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""

Parsing and validation of admixture results.

"""

import collections

CUTOFF = 0.001


class AdmixtureError(RuntimeError):
    pass


def read_admixture_results(filename, data, k_groups, cutoff=CUTOFF):
    names = tuple(data.sample_order) + ("-",)
    table = _admixture_read_results(filename, names)
    _admixture_validate_ancestral_groups(data, table, k_groups, cutoff)

    ancestral_groups = [[set(), value] for value in table["-"]]
    for sample, row in table.items():
        if sample == "-":
            continue

        group = data.groups[k_groups][sample]
        for index, value in enumerate(row):
            if value >= cutoff:
                ancestral_groups[index][0].add(group)

    return ancestral_groups


def get_percentiles(data, sample1, sample2, nreads, k_groups, has_ts, value):
    results = {"Sample1": sample1, "Sample2": sample2}

    nreads_lower = set(
        row["NReads"] for row in data.simulations if row["NReads"] <= nreads
    )
    nreads_upper = set(
        row["NReads"] for row in data.simulations if row["NReads"] >= nreads
    )

    if nreads_lower:
        selection = _select_simulations(
            data=data,
            sample1=sample1,
            sample2=sample2,
            nreads=max(nreads_lower),
            k_groups=k_groups,
            has_ts=has_ts,
        )
        lower_bound, upper_bound = _get_percentile_range(selection, value)
        results["Lower"] = {
            "NReads": max(nreads_lower),
            "Lower": lower_bound,
            "Upper": upper_bound,
        }

    if nreads_upper:
        selection = _select_simulations(
            data=data,
            sample1=sample1,
            sample2=sample2,
            nreads=min(nreads_upper),
            k_groups=k_groups,
            has_ts=has_ts,
        )
        lower_bound, upper_bound = _get_percentile_range(selection, value)
        results["Upper"] = {
            "NReads": min(nreads_upper),
            "Lower": lower_bound,
            "Upper": upper_bound,
        }

    return results


def _select_simulations(data, sample1, sample2, nreads, k_groups, has_ts):
    selection = []
    samples = frozenset((sample1, sample2))
    for row in data.simulations:
        if row["K"] != k_groups or row["HasTS"] != has_ts:
            continue
        elif row["NReads"] != nreads:
            continue
        elif frozenset((row["Sample1"], row["Sample2"])) != samples:
            continue

        selection.append(row)

    return selection


def _get_percentile_range(selection, value):
    selection = [(row["Percentile"], row["Value"]) for row in selection]
    selection.sort()

    lower_bound = 0.0
    upper_bound = 1.0

    for cur_pct, cur_value in selection:
        if cur_value > value:
            break

        lower_bound = cur_pct

    for cur_pct, cur_value in reversed(selection):
        if cur_value < value:
            break

        upper_bound = cur_pct

    return lower_bound, upper_bound


def _admixture_read_results(filename, samples):
    with open(filename) as handle:
        lines = handle.readlines()

    if len(samples) != len(lines):
        raise AdmixtureError(
            "unexpected number of lines in admixture file; "
            "expected %i samples, found %i" % (len(samples), len(lines))
        )

    result = {}
    for name, line in zip(samples, lines):
        result[name] = [float(value) for value in line.split()]

    return result


def _admixture_validate_ancestral_groups(data, table, k_groups, cutoff):
    groups = collections.defaultdict(dict)
    for sample, row in table.items():
        group = data.groups[k_groups].get(sample)
        if group is not None:
            for index, value in enumerate(row):
                if value >= cutoff:
                    groups[group][index] = True

    mixed_groups = []
    for group, memberships in sorted(groups.items()):
        count = len(memberships)

        if count > 1:
            mixed_groups.append(
                "member(s) of reference group %s assigned to "
                "%i ancestral populations" % (group, count)
            )

    if mixed_groups:
        raise AdmixtureError(
            "Inconsistent ADMIXTURE results: %s; "
            "cannot determine ancestry!" % ("; ".join(mixed_groups))
        )
